"""
MIT License

Copyright (c) 2019 Conor Durkan, Artur Bekasov, Iain Murray, George Papamakarios

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Implementations of invertible non-linearities.

Slightly modified from:
https://github.com/bayesiains/nsf/blob/master/nde/transforms/nonlinearities.py
"""
# pylint:disable=missing-class-docstring
import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from ray.rllib.utils import override

from .abstract import CompositeTransform
from .abstract import InverseTransform
from .abstract import Transform
from .splines import DEFAULT_MIN_BIN_HEIGHT
from .splines import DEFAULT_MIN_BIN_WIDTH
from .splines import DEFAULT_MIN_DERIVATIVE
from .splines import unconstrained_rational_quadratic_spline
from .utils import sum_rightmost


class Tanh(Transform):
    def encode(self, inputs):
        outputs = torch.tanh(inputs)
        logabsdet = torch.log(1 - outputs ** 2)
        logabsdet = sum_rightmost(logabsdet, self.event_dim)
        return outputs, logabsdet

    def decode(self, inputs):
        outputs = 0.5 * torch.log((1 + inputs) / (1 - inputs))
        logabsdet = -torch.log(1 - inputs ** 2)
        logabsdet = sum_rightmost(logabsdet, self.event_dim)
        return outputs, logabsdet


class LogTanh(Transform):
    """
    Tanh with unbounded output. Constructed by selecting a cut_point, and replacing
    values to the right of cut_point with alpha * log(beta * x), and to the left of
    -cut_point with -alpha * log(-beta * x). alpha and beta are set to match the value
    and the first derivative of tanh at cut_point.
    """

    def __init__(self, cut_point=1):
        if cut_point <= 0:
            raise ValueError("Cut point must be positive.")
        super().__init__()

        self.cut_point = cut_point
        self.inv_cut_point = np.tanh(cut_point)

        self.alpha = (1 - np.tanh(np.tanh(cut_point))) / cut_point
        self.beta = np.exp(
            (np.tanh(cut_point) - self.alpha * np.log(cut_point)) / self.alpha
        )

    def encode(self, inputs):
        mask_right = inputs > self.cut_point
        mask_left = inputs < -self.cut_point
        mask_middle = ~(mask_right | mask_left)

        outputs = torch.zeros_like(inputs)
        outputs[mask_middle] = torch.tanh(inputs[mask_middle])
        outputs[mask_right] = self.alpha * torch.log(self.beta * inputs[mask_right])
        outputs[mask_left] = self.alpha * -torch.log(-self.beta * inputs[mask_left])

        logabsdet = torch.zeros_like(inputs)
        logabsdet[mask_middle] = torch.log(1 - outputs[mask_middle] ** 2)
        logabsdet[mask_right] = torch.log(self.alpha / inputs[mask_right])
        logabsdet[mask_left] = torch.log(-self.alpha / inputs[mask_left])
        logabsdet = sum_rightmost(logabsdet, self.event_dim)

        return outputs, logabsdet

    def decode(self, inputs):

        mask_right = inputs > self.inv_cut_point
        mask_left = inputs < -self.inv_cut_point
        mask_middle = ~(mask_right | mask_left)

        outputs = torch.zeros_like(inputs)
        outputs[mask_middle] = 0.5 * torch.log(
            (1 + inputs[mask_middle]) / (1 - inputs[mask_middle])
        )
        outputs[mask_right] = torch.exp(inputs[mask_right] / self.alpha) / self.beta
        outputs[mask_left] = -torch.exp(-inputs[mask_left] / self.alpha) / self.beta

        logabsdet = torch.zeros_like(inputs)
        logabsdet[mask_middle] = -torch.log(1 - inputs[mask_middle] ** 2)
        logabsdet[mask_right] = (
            -np.log(self.alpha * self.beta) + inputs[mask_right] / self.alpha
        )
        logabsdet[mask_left] = (
            -np.log(self.alpha * self.beta) - inputs[mask_left] / self.alpha
        )
        logabsdet = sum_rightmost(logabsdet, self.event_dim)

        return outputs, logabsdet


class LeakyReLU(Transform):
    def __init__(self, negative_slope=1e-2):
        if negative_slope <= 0:
            raise ValueError("Slope must be positive.")
        super().__init__()
        self.negative_slope = negative_slope
        self.log_negative_slope = torch.log(torch.as_tensor(self.negative_slope))

    def encode(self, inputs):
        outputs = F.leaky_relu(inputs, negative_slope=self.negative_slope)
        mask = (inputs < 0).type(torch.Tensor)
        logabsdet = self.log_negative_slope * mask
        logabsdet = sum_rightmost(logabsdet, self.event_dim)
        return outputs, logabsdet

    def decode(self, inputs):
        outputs = F.leaky_relu(inputs, negative_slope=(1 / self.negative_slope))
        mask = (inputs < 0).type(torch.Tensor)
        logabsdet = -self.log_negative_slope * mask
        logabsdet = sum_rightmost(logabsdet, self.event_dim)
        return outputs, logabsdet


class Sigmoid(Transform):
    def __init__(self, temperature=1, eps=1e-6):
        super().__init__()
        self.eps = eps
        self.temperature = torch.Tensor([temperature])

    def encode(self, inputs):
        inputs = self.temperature * inputs
        outputs = torch.sigmoid(inputs)
        logabsdet = sum_rightmost(
            torch.log(self.temperature) - F.softplus(-inputs) - F.softplus(inputs),
            self.event_dim,
        )
        return outputs, logabsdet

    def decode(self, inputs):
        if torch.min(inputs) < 0 or torch.max(inputs) > 1:
            raise ValueError("Input outside domain")

        inputs = torch.clamp(inputs, self.eps, 1 - self.eps)

        outputs = (1 / self.temperature) * (torch.log(inputs) - torch.log1p(-inputs))
        logabsdet = -sum_rightmost(
            torch.log(self.temperature)
            - F.softplus(-self.temperature * outputs)
            - F.softplus(self.temperature * outputs),
            self.event_dim,
        )
        return outputs, logabsdet


class Logit(InverseTransform):
    def __init__(self, temperature=1, eps=1e-6):
        super().__init__(Sigmoid(temperature=temperature, eps=eps))


class CauchyCDF(Transform):
    def encode(self, inputs):
        outputs = (1 / np.pi) * torch.atan(inputs) + 0.5
        logabsdet = sum_rightmost(
            -np.log(np.pi) - torch.log(1 + inputs ** 2), self.event_dim
        )
        return outputs, logabsdet

    def decode(self, inputs):
        if torch.min(inputs) < 0 or torch.max(inputs) > 1:
            raise ValueError("Input outside domain")

        outputs = torch.tan(np.pi * (inputs - 0.5))
        logabsdet = -sum_rightmost(
            -np.log(np.pi) - torch.log(1 + outputs ** 2), self.event_dim
        )
        return outputs, logabsdet


class CauchyCDFInverse(InverseTransform):
    def __init__(self):
        super().__init__(CauchyCDF())


class CompositeCDFTransform(CompositeTransform):
    def __init__(self, squashing_transform, cdf_transform):
        super().__init__(
            [squashing_transform, cdf_transform, InverseTransform(squashing_transform)]
        )


class PiecewiseRationalQuadraticCDF(Transform):
    # pylint:disable=too-many-instance-attributes
    def __init__(
        self,
        shape,
        num_bins=10,
        tail_bound=1.0,
        identity_init=False,
        min_bin_width=DEFAULT_MIN_BIN_WIDTH,
        min_bin_height=DEFAULT_MIN_BIN_HEIGHT,
        min_derivative=DEFAULT_MIN_DERIVATIVE,
        event_dim=1,
    ):
        # pylint:disable=too-many-arguments
        super().__init__(event_dim=event_dim)

        self.min_bin_width = min_bin_width
        self.min_bin_height = min_bin_height
        self.min_derivative = min_derivative

        self.tail_bound = tail_bound

        if identity_init:
            self.unnormalized_widths = nn.Parameter(torch.zeros(*shape, num_bins))
            self.unnormalized_heights = nn.Parameter(torch.zeros(*shape, num_bins))
            constant = np.log(np.exp(1 - min_derivative) - 1)
            self.unnormalized_derivatives = nn.Parameter(
                torch.empty(*shape, num_bins - 1).fill_(constant)
            )
        else:
            self.unnormalized_widths = nn.Parameter(torch.rand(*shape, num_bins))
            self.unnormalized_heights = nn.Parameter(torch.rand(*shape, num_bins))
            self.unnormalized_derivatives = nn.Parameter(
                torch.rand(*shape, num_bins - 1)
            )

    def _spline(self, inputs, inverse: bool = False):
        batch_shape = inputs.shape[: -self.event_dim]
        unnormalized_widths = self.unnormalized_widths.expand(
            batch_shape + self.unnormalized_widths.shape
        )
        unnormalized_heights = self.unnormalized_heights.expand(
            batch_shape + self.unnormalized_heights.shape
        )
        unnormalized_derivatives = self.unnormalized_derivatives.expand(
            batch_shape + self.unnormalized_derivatives.shape
        )

        # Always use linear tails
        outputs, logabsdet = unconstrained_rational_quadratic_spline(
            inputs=inputs,
            unnormalized_widths=unnormalized_widths,
            unnormalized_heights=unnormalized_heights,
            unnormalized_derivatives=unnormalized_derivatives,
            inverse=inverse,
            tail_bound=self.tail_bound,
            min_bin_width=self.min_bin_width,
            min_bin_height=self.min_bin_height,
            min_derivative=self.min_derivative,
        )

        return outputs, sum_rightmost(logabsdet, self.event_dim)

    def encode(self, inputs):
        return self._spline(inputs, inverse=False)

    def decode(self, inputs):
        return self._spline(inputs, inverse=True)


##################################################################
# Custom
##################################################################


class TanhTransform(Transform):
    """Transform via the mapping :math:`y = \frac{e^x - e^{-x}} {e^x + e^{-x}}`."""

    # pylint:disable=arguments-out-of-order

    @override(Transform)
    def encode(self, inputs):
        outputs = torch.tanh(inputs)
        return outputs, self.log_abs_det_jacobian(inputs, outputs)

    @override(Transform)
    def decode(self, inputs):
        # torch.finfo(torch.float32).tiny
        to_log1 = torch.clamp(1 + inputs, min=1.1754943508222875e-38)
        to_log2 = torch.clamp(1 - inputs, min=1.1754943508222875e-38)
        outputs = (torch.log(to_log1) - torch.log(to_log2)) / 2
        return outputs, -self.log_abs_det_jacobian(outputs, inputs)

    def log_abs_det_jacobian(self, inputs, outputs):
        # pylint:disable=unused-argument,missing-docstring
        # Taken from spinningup's implementation of SAC
        return sum_rightmost(
            2 * (math.log(2) - inputs - F.softplus(-2 * inputs)), self.event_dim
        )


class SigmoidTransform(Transform):
    # pylint:disable=missing-docstring,arguments-out-of-order

    @override(Transform)
    def encode(self, inputs):
        outputs = inputs.sigmoid()
        return outputs, self.log_abs_det_jacobian(inputs, outputs)

    @override(Transform)
    def decode(self, inputs):
        to_log = inputs.clamp(min=1.1754943508222875e-38)
        outputs = to_log.log() - (-to_log).log1p()
        return outputs, -self.log_abs_det_jacobian(outputs, inputs)

    def log_abs_det_jacobian(self, inputs, outputs):
        # pylint:disable=unused-argument,missing-docstring
        return sum_rightmost(-F.softplus(-inputs) - F.softplus(inputs), self.event_dim)


class AffineTransform(Transform):
    # pylint:disable=missing-docstring,arguments-out-of-order

    def __init__(self, loc, scale, **kwargs):
        super().__init__(**kwargs)
        self.register_buffer("loc", loc)
        self.register_buffer("scale", scale)

    @override(Transform)
    def encode(self, inputs):
        outputs = inputs * self.scale + self.loc
        return outputs, self.log_abs_det_jacobian(inputs, outputs)

    @override(Transform)
    def decode(self, inputs):
        outputs = (inputs - self.loc) / self.scale
        return outputs, -self.log_abs_det_jacobian(outputs, inputs)

    def log_abs_det_jacobian(self, inputs, outputs):
        # pylint:disable=unused-argument,missing-docstring
        _, scale = torch.broadcast_tensors(inputs, self.scale)
        return sum_rightmost(scale.abs().log(), self.event_dim)


class TanhSquashTransform(Transform):
    """Squashes samples to the desired range using Tanh."""

    def __init__(self, low, high, event_dim=0):
        divide = AffineTransform(loc=torch.zeros_like(low), scale=2 / (high - low))
        squash = TanhTransform()
        shift = AffineTransform(loc=(high + low) / 2, scale=(high - low) / 2)
        compose = CompositeTransform([divide, squash, shift], event_dim=event_dim)
        super().__init__(cond_transform=compose)


class SigmoidSquashTransform(Transform):
    """Squashes samples to the desired range using Sigmoid."""

    def __init__(self, low, high, event_dim=0):
        divide = AffineTransform(loc=torch.zeros_like(low), scale=1 / (high - low))
        squash = SigmoidTransform()
        shift = AffineTransform(loc=low, scale=(high - low))
        compose = CompositeTransform([divide, squash, shift], event_dim=event_dim)
        super().__init__(cond_transform=compose)



"""Trainer and default config for MAGE."""
from raylab.agents import trainer
from raylab.agents.model_based import ModelBasedTrainer
from raylab.agents.sop.trainer import sop_config
from raylab.policy.model_based.training_mixin import TrainingSpec

from .policy import MAGETorchPolicy


@trainer.configure
@trainer.option("lambda", default=0.05, help="TD error regularization for MAGE loss")
@trainer.option(
    "model_training", default=TrainingSpec().to_dict(), help=TrainingSpec.__doc__
)
@trainer.option(
    "model_warmup",
    default=TrainingSpec().to_dict(),
    help="""Specifications for model warm-up.

    Same configurations as 'model_training'.
    """,
)
@sop_config
@trainer.option("module/type", "ModelBasedDDPG")
@trainer.option("torch_optimizer/models", {"type": "Adam"})
@trainer.option(
    "exploration_config/type", "raylab.utils.exploration.GaussianNoise", override=True
)
@trainer.option("holdout_ratio", default=0, override=True)
@trainer.option("max_holdout", default=0, override=True)
@trainer.option("evaluation_config/explore", False, override=True)
class MAGETrainer(ModelBasedTrainer):
    """Single agent trainer for MAGE."""

    _name = "MAGE"
    _policy = MAGETorchPolicy



# pylint:disable=missing-module-docstring
from raylab.agents.sac import SACTorchPolicy
from raylab.policy import EnvFnMixin
from raylab.policy.losses import DAPO


class DAPOTorchPolicy(EnvFnMixin, SACTorchPolicy):
    """Dynamics-Aware Policy Optimization policy in PyTorch to use with RLlib.

    Effectively substitutes the actor loss in SAC for a dynamics-aware 1-step
    approximation of the expected Q-value.
    """

    # pylint:disable=abstract-method,attribute-defined-outside-init

    @property
    def options(self):
        """Return the default configuration for MAPO."""
        # pylint:disable=cyclic-import
        from raylab.agents.mapo.dapo import DAPOTrainer

        return DAPOTrainer.options

    def _setup_actor_loss(self):
        # Can only be defined once we have access to the env dynamics
        self.loss_actor = None

    def _set_reward_hook(self):
        if self.loss_actor:
            self.loss_actor.set_reward_fn(self.reward_fn)

    def _set_termination_hook(self):
        if self.loss_actor:
            self.loss_actor.set_termination_fn(self.termination_fn)

    def _set_dynamics_hook(self):
        self.loss_actor = DAPO(self.dynamics_fn, self.module.actor, self.module.critics)
        self.loss_actor.gamma = self.config["gamma"]
        self.loss_actor.dynamics_samples = self.config["losses"]["model_samples"]
        self.loss_actor.grad_estimator = self.config["losses"]["grad_estimator"]

        if self.reward_fn:
            self.loss_actor.set_reward_fn(self.reward_fn)
        if self.termination_fn:
            self.loss_actor.set_termination_fn(self.termination_fn)



"""Model-based Actor-Critic for the maximum entropy framework."""

from .abstract import AbstractModelActorCritic
from .mixins import MaximumEntropyMixin
from .mixins import StateValueMixin
from .mixins import StochasticActorMixin
from .mixins import StochasticModelMixin


# pylint:disable=abstract-method,too-many-ancestors
class MaxEntModelBased(
    StochasticModelMixin,
    StochasticActorMixin,
    StateValueMixin,
    MaximumEntropyMixin,
    AbstractModelActorCritic,
):
    """
    Module architecture with stochastic actor and model, state-value function, and
    entropy coefficient.
    """



"""Trainer and configuration for MAPO with maximum likelihood-trained model."""
from raylab.agents import trainer
from raylab.agents.model_based import ModelBasedTrainer
from raylab.agents.sac.trainer import sac_config
from raylab.policy.model_based.training_mixin import TrainingSpec

from .policy import MlMAPOTorchPolicy


@trainer.configure
@trainer.option("losses/", help="Configurations for model and actor loss functions")
@trainer.option(
    "losses/grad_estimator",
    default="PD",
    help="""Gradient estimator for optimizing expectations.

    Possible types include
    SF: score function
    PD: pathwise derivative
    """,
)
@trainer.option(
    "losses/model_samples",
    default=1,
    help="Number of next states to sample from the model when calculating the"
    " model-aware deterministic policy gradient",
)
@trainer.option("module/type", default="ModelBasedSAC")
@trainer.option("torch_optimizer/models/type", "Adam")
@trainer.option(
    "model_training", default=TrainingSpec().to_dict(), help=TrainingSpec.__doc__
)
@trainer.option(
    "model_warmup",
    default=TrainingSpec().to_dict(),
    help="""Specifications for model warm-up.

    Same configurations as 'model_training'.
    """,
)
@trainer.option("evaluation_config/explore", False, override=True)
@trainer.option("rollout_fragment_length", 25, override=True)
@sac_config
class MlMAPOTrainer(ModelBasedTrainer):
    """Single agent trainer for MAPO-MLE."""

    _name = "MAPO-MLE"
    _policy = MlMAPOTorchPolicy



"""Actor-Critic architecture used in Soft Actor-Critic (SAC)."""
from raylab.utils.dictionaries import deep_merge

from .abstract import AbstractActorCritic
from .mixins import ActionValueMixin
from .mixins import MaximumEntropyMixin
from .mixins import StochasticActorMixin


BASE_CONFIG = {
    "actor": {
        "encoder": {
            "units": (400, 300),
            "activation": "ReLU",
            "initializer_options": {"name": "xavier_uniform"},
        },
        "input_dependent_scale": True,
    },
    "critic": {
        "double_q": False,
        "encoder": {
            "units": (400, 300),
            "activation": "ReLU",
            "initializer_options": {"name": "xavier_uniform"},
            "delay_action": True,
        },
    },
    "entropy": {"initial_alpha": 0.05},
}


class SACModule(
    StochasticActorMixin, ActionValueMixin, MaximumEntropyMixin, AbstractActorCritic
):
    """Actor-Critic module with stochastic actor and action-value critics."""

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        config = deep_merge(BASE_CONFIG, config, False, ["actor", "critic", "entropy"])
        super().__init__(obs_space, action_space, config)



"""Actor-Critic architecture popularized by DDPG."""
from raylab.utils.dictionaries import deep_merge

from .abstract import AbstractActorCritic
from .mixins import ActionValueMixin
from .mixins import DeterministicActorMixin


BASE_CONFIG = {
    "actor": {
        # === Twin Delayed DDPG (TD3) tricks ===
        # Add gaussian noise to the action when calculating the target Q function
        "smooth_target_policy": True,
        # Additive Gaussian i.i.d. noise to add to actions inputs to target Q function
        "target_gaussian_sigma": 0.3,
        "separate_target_policy": False,
        "perturbed_policy": False,
        # === SQUASHING EXPLORATION PROBLEM ===
        # Maximum l1 norm of the policy's output vector before the squashing
        # function
        "beta": 1.2,
        "encoder": {
            "units": (400, 300),
            "activation": "ReLU",
            "initializer_options": {"name": "xavier_uniform"},
            "layer_norm": False,
        },
    },
    "critic": {
        "double_q": False,
        "encoder": {
            "units": (400, 300),
            "activation": "ReLU",
            "initializer_options": {"name": "xavier_uniform"},
            "delay_action": True,
        },
    },
}


class DDPGModule(DeterministicActorMixin, ActionValueMixin, AbstractActorCritic):
    """Actor-Critic module with deterministic actor and action-value critics."""

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        config = deep_merge(BASE_CONFIG, config, False, ["actor", "critic"])
        super().__init__(obs_space, action_space, config)



"""Trainer and configuration for SOP."""
from raylab.agents import trainer
from raylab.agents.off_policy import OffPolicyTrainer

from .policy import SOPTorchPolicy


def sop_config(cls: type) -> type:
    """Add configurations for Streamlined Off-Policy-based agents."""
    for config_setter in [
        trainer.option(
            "dpg_loss",
            "default",
            help="""
            Type of Deterministic Policy Gradient to use.

            'default' backpropagates Q-value gradients through the critic network.

            'acme' uses Acme's implementation which recovers DPG via a MSE loss between
            the actor's action and the action + Q-value gradient. Allows monitoring the
            magnitude of the action-value gradient.""",
        ),
        trainer.option(
            "dqda_clipping",
            None,
            help="""
            Optional value by which to clip the action gradients. Only used with
            dpg_loss='acme'.""",
        ),
        trainer.option(
            "clip_dqda_norm",
            False,
            help="""
            Whether to clip action grads by norm or value. Only used with
            dpg_loss='acme'.""",
        ),
        trainer.option("torch_optimizer/actor", {"type": "Adam", "lr": 1e-3}),
        trainer.option("torch_optimizer/critics", {"type": "Adam", "lr": 1e-3}),
        trainer.option(
            "polyak",
            0.995,
            help="Interpolation factor in polyak averaging for target networks.",
        ),
        trainer.option(
            "policy_delay",
            1,
            help="Update policy every this number of calls to `learn_on_batch`",
        ),
    ]:
        cls = config_setter(cls)

    return cls


@trainer.configure
@sop_config
@trainer.option("module/type", "DDPG")
@trainer.option("module/actor/separate_behavior", True)
@trainer.option(
    "exploration_config/type", "raylab.utils.exploration.ParameterNoise", override=True
)
@trainer.option(
    "exploration_config/param_noise_spec",
    {"initial_stddev": 0.1, "desired_action_stddev": 0.2, "adaptation_coeff": 1.01},
    help="Options for parameter noise exploration",
)
@trainer.option("exploration_config/pure_exploration_steps", 1000)
@trainer.option("evaluation_config/explore", False, override=True)
class SOPTrainer(OffPolicyTrainer):
    """Single agent trainer for the Streamlined Off-Policy algorithm."""

    _name = "SOP"
    _policy = SOPTorchPolicy



"""Registry of old modules for PyTorch policies."""
import torch.nn as nn
from gym.spaces import Space

from .maxent_model_based import MaxEntModelBased
from .nfmbrl import NFMBRL
from .off_policy_nfac import OffPolicyNFAC
from .on_policy_actor_critic import OnPolicyActorCritic
from .on_policy_nfac import OnPolicyNFAC
from .simple_model_based import SimpleModelBased
from .svg_module import SVGModule
from .svg_realnvp_actor import SVGRealNVPActor
from .trpo_tang2018 import TRPOTang2018

MODULES = {
    cls.__name__ + "-v0": cls
    for cls in (
        SimpleModelBased,
        SVGModule,
        MaxEntModelBased,
        NFMBRL,
        OnPolicyActorCritic,
        OnPolicyNFAC,
        OffPolicyNFAC,
        TRPOTang2018,
        SVGRealNVPActor,
    )
}


def get_module(obs_space: Space, action_space: Space, config: dict) -> nn.Module:
    """Retrieve and construct module of given name.

    Args:
        obs_space: Observation space
        action_space: Action space
        config: Configurations for module construction and initialization
    """
    type_ = config.pop("type")
    return MODULES[type_](obs_space, action_space, config)



"""Trainer and configuration for SVG(inf)."""
from ray.rllib import SampleBatch
from ray.rllib.utils import override

from raylab.agents import Trainer
from raylab.agents import trainer
from raylab.agents.model_based import set_policy_with_env_fn
from raylab.utils.replay_buffer import NumpyReplayBuffer
from raylab.utils.replay_buffer import ReplayField

from .policy import SVGInfTorchPolicy


@trainer.configure
@trainer.option(
    "vf_loss_coeff",
    1.0,
    help="Weight of the fitted V loss in the joint model-value loss",
)
@trainer.option("max_grad_norm", 10.0, help="Clip gradient norms by this value")
@trainer.option(
    "max_is_ratio", 5.0, help="Clip importance sampling weights by this value"
)
@trainer.option(
    "polyak",
    0.995,
    help="Interpolation factor in polyak averaging for target networks.",
)
@trainer.option("torch_optimizer/on_policy", {"type": "Adam", "lr": 1e-3})
@trainer.option("torch_optimizer/off_policy", {"type": "Adam", "lr": 1e-3})
@trainer.option(
    "updates_per_step",
    1.0,
    help="Model and Value function updates per step in the environment",
)
@trainer.option("buffer_size", 500000, help="Size of the replay buffer")
@trainer.option(
    "kl_schedule/",
    help="Options for adaptive KL coefficient. See raylab.utils.adaptive_kl",
    allow_unknown_subkeys=True,
)
@trainer.option("module", {"type": "SVGModule-v0"}, override=True)
@trainer.option(
    "exploration_config/type", "raylab.utils.exploration.StochasticActor", override=True
)
@trainer.option("evaluation_config/explore", True)
@trainer.option("num_workers", 0, override=True)
@trainer.option("rollout_fragment_length", 1, override=True)
@trainer.option("batch_mode", "complete_episodes", override=True)
@trainer.option("train_batch_size", 128, override=True)
class SVGInfTrainer(Trainer):
    """Single agent trainer for SVG(inf)."""

    # pylint:disable=attribute-defined-outside-init

    _name = "SVG(inf)"
    _policy = SVGInfTorchPolicy

    @override(Trainer)
    def _init(self, config, env_creator):
        self._validate_config(config)
        self.workers = self._make_workers(
            env_creator, self._policy, config, num_workers=config["num_workers"]
        )
        set_policy_with_env_fn(self.workers, fn_type="reward")

        policy = self.get_policy()
        self.replay = NumpyReplayBuffer(
            policy.observation_space, policy.action_space, config["buffer_size"]
        )
        self.replay.add_fields(ReplayField(SampleBatch.ACTION_LOGP))
        self.replay.seed(config["seed"])

    @override(Trainer)
    def _train(self):
        init_timesteps = self.metrics.num_steps_sampled
        worker = self.workers.local_worker()
        policy = worker.get_policy()

        samples = worker.sample()
        self.metrics.num_steps_sampled += samples.count
        for row in samples.rows():
            self.replay.add(row)
        stats = policy.get_exploration_info()

        with policy.learning_off_policy():
            for _ in range(int(samples.count * self.config["updates_per_step"])):
                batch = self.replay.sample(self.config["train_batch_size"])
                off_policy_stats = policy.learn_on_batch(batch)
                self.metrics.num_steps_trained += batch.count
        stats.update(off_policy_stats)

        stats.update(policy.learn_on_batch(samples))

        timesteps_this_iter = self.metrics.num_steps_sampled - init_timesteps
        return self._log_metrics(stats, timesteps_this_iter)

    def _log_metrics(self, learner_stats, timesteps_this_iter):
        res = self.collect_metrics()
        res.update(timesteps_this_iter=timesteps_this_iter, learner=learner_stats)
        return res



"""Support for modules with deterministic policies."""
import warnings

import torch
import torch.nn as nn
from ray.rllib.utils import override

from raylab.pytorch.nn import FullyConnected
from raylab.pytorch.nn import GaussianNoise
from raylab.pytorch.nn import NormalizedLinear
from raylab.pytorch.nn import TanhSquash
from raylab.pytorch.nn.init import initialize_
from raylab.utils.dictionaries import deep_merge


BASE_CONFIG = {
    # === Twin Delayed DDPG (TD3) tricks ===
    # Add gaussian noise to the action when calculating the Deterministic
    # Policy Gradient
    "smooth_target_policy": True,
    # Additive Gaussian i.i.d. noise to add to actions inputs to target Q function
    "target_gaussian_sigma": 0.3,
    "separate_target_policy": False,
    "perturbed_policy": False,
    # === SQUASHING EXPLORATION PROBLEM ===
    # Maximum l1 norm of the policy's output vector before the squashing
    # function
    "beta": 1.2,
    "initializer_options": {"name": "xavier_uniform"},
    "encoder": {"units": (32, 32), "activation": "ReLU", "layer_norm": False},
}


class DeterministicActorMixin:
    """Adds constructor for modules with deterministic policies."""

    # pylint:disable=too-few-public-methods

    @staticmethod
    def _make_actor(obs_space, action_space, config):
        config = deep_merge(BASE_CONFIG, config.get("actor", {}), False, ["encoder"])
        if not config["encoder"].get("layer_norm") and config["perturbed_policy"]:
            warnings.warn(
                "'layer_norm' is deactivated even though a perturbed policy was "
                "requested. For optimal stability, set 'layer_norm': True."
            )

        actor = DeterministicPolicy.from_scratch(obs_space, action_space, config)

        behavior = actor
        if config["perturbed_policy"]:
            behavior = DeterministicPolicy.from_scratch(obs_space, action_space, config)

        target_actor = actor
        if config["separate_target_policy"]:
            target_actor = DeterministicPolicy.from_scratch(
                obs_space, action_space, config
            )
            target_actor.load_state_dict(actor.state_dict())
        if config["smooth_target_policy"]:
            target_actor = DeterministicPolicy.from_existing(
                target_actor, noise=config["target_gaussian_sigma"],
            )

        return {"actor": actor, "behavior": behavior, "target_actor": target_actor}


class DeterministicPolicy(nn.Module):
    """Represents a deterministic policy as a sequence of modules.

    If a noise param is passed, adds Gaussian white noise before the squashing module.
    """

    def __init__(self, mods, *, noise=None):
        super().__init__()
        if noise:
            noise_mod = nn.ModuleList([GaussianNoise(noise)])
            mods = mods[:-1].extend(noise_mod).extend(mods[-1:])
        self.mods = mods
        self.sequential = nn.Sequential(*self.mods)

    @classmethod
    def from_existing(cls, policy, *, noise=None):
        """Create a policy using an existing's modules."""
        mods = policy.mods
        if len(mods) > 3:
            mods = mods[:2].extend(mods[-1:])
        return cls(mods, noise=noise)

    @classmethod
    def from_scratch(cls, obs_space, action_space, config, *, noise=None):
        """Create a policy using new modules."""
        logits = FullyConnected(in_features=obs_space.shape[0], **config["encoder"])
        logits.apply(
            initialize_(
                activation=config["encoder"].get("activation"),
                **config["initializer_options"]
            )
        )
        normalize = NormalizedLinear(
            in_features=logits.out_features,
            out_features=action_space.shape[0],
            beta=config["beta"],
        )
        squash = TanhSquash(
            torch.as_tensor(action_space.low), torch.as_tensor(action_space.high),
        )
        mods = nn.ModuleList([logits, normalize, squash])
        return cls(mods, noise=noise)

    @override(nn.Module)
    def forward(self, obs):  # pylint:disable=arguments-differ
        return self.sequential(obs)



"""Primitives for all Trainers."""
import warnings
from abc import ABCMeta
from typing import Callable
from typing import Optional

try:
    import wandb
except ImportError:
    wandb = None

from ray.rllib.agents.trainer import Trainer as RLlibTrainer
from ray.rllib.evaluation.worker_set import WorkerSet
from ray.rllib.optimizers import PolicyOptimizer
from ray.rllib.utils import override as overrides
from ray.tune import Trainable

from . import compat
from .options import Json
from .options import RaylabOptions


# ==============================================================================
# Programatic config setting
# ==============================================================================
def configure(cls: type) -> type:
    """Decorator for finishing the configuration setup for a Trainer class.

    Should be called after all :func:`config` decorators have been applied,
    i.e., as the top-most decorator.
    """
    cls.options = cls.options.copy_and_set_queued_options()
    return cls


def option(
    key: str,
    default: Json = None,
    *,
    help: Optional[str] = None,
    override: bool = False,
    allow_unknown_subkeys: bool = False,
    override_all_if_type_changes: bool = False,
    separator: str = "/",
) -> Callable[[type], type]:
    """Returns a decorator for adding/overriding a Trainer class config.

    If `key` ends in a separator and `default` is None, treats the option as a
    nested dict of options and sets the default to an empty dictionary.

    Args:
        key: Name of the config paremeter which the use can tune
        default: Default Jsonable value to set for the parameter
        info: Parameter help text explaining what the parameter does
        override: Whether to override an existing parameter
        allow_unknown_subkeys: Whether to allow new keys for dict parameters.
            This is only at the top level
        override_all_if_type_changes: Whether to override the entire value
            (dict) iff the 'type' key in this value dict changes. This is only
            at the top level
        separator: String token separating nested keys

    Raises:
        RuntimeError: If attempting to set an existing parameter with `override`
            set to `False`.
    """
    # pylint:disable=too-many-arguments,redefined-builtin
    def _queue(cls):
        cls.options.add_option_to_queue(
            key=key,
            default=default,
            info=help,
            override=override,
            allow_unknown_subkeys=allow_unknown_subkeys,
            override_all_if_type_changes=override_all_if_type_changes,
            separator=separator,
        )
        return cls

    return _queue


# ==============================================================================
# Base Raylab Trainer
# ==============================================================================
Jsonable = (dict, list, str, int, float, bool, type(None))


@configure
@option(
    "wandb/",
    allow_unknown_subkeys=True,
    help="""Configs for integration with Weights & Biases.

    Accepts arbitrary keyword arguments to pass to `wandb.init`.
    The defaults for `wandb.init` are:
    * name: `_name` property of the trainer.
    * config: full `config` attribute of the trainer
    * config_exclude_keys: `wandb` and `callbacks` configs
    * reinit: True

    Don't forget to:
      * install `wandb` via pip
      * login to W&B with the appropriate API key for your
        team/project.
      * set the `wandb/project` name in the config dict

    Check out the Quickstart for more information:
    `https://docs.wandb.com/quickstart`
    """,
)
@option(
    "wandb/enabled",
    default=False,
    help="""Whether to sync logs and files with `wandb`.""",
)
@option(
    "wandb/file_paths",
    default=(),
    help="Sequence of file names to pass to `wandb.save` on trainer setup.",
)
@option(
    "wandb/save_checkpoints",
    default=False,
    help="Whether to sync trainer checkpoints to W&B via `wandb.save`.",
)
@option("compile_policy", False, help="Whether to optimize the policy's backend")
@option(
    "module/",
    help="Type and config of the PyTorch NN module.",
    allow_unknown_subkeys=True,
    override_all_if_type_changes=True,
)
@option(
    "torch_optimizer/",
    help="Config dict for PyTorch optimizers.",
    allow_unknown_subkeys=True,
)
class Trainer(RLlibTrainer, metaclass=ABCMeta):
    """Base Trainer for all agents.

    Either a PolicyOptimizer or WorkerSet must be set (as `optimizer` or
    `workers` attributes respectively) to collect episode statistics.

    Always creates a PolicyOptimizer instance as the `optimizer` attribute to
    log episode metrics (to be removed in the future).

    Accessing `metrics` returns the optimizer, which is useful when using an
    optimizer only to log metrics. This way, the user can update
    `self.metrics.num_steps_sampled` and the results will be logged via RLlib's
    framework.

    Integration with `Weights & Biases`_ is available. The user must install
    `wandb` and set a project name in the `wandb` subconfig dict.

    .. _`Weights & Biases`: https://docs.wandb.com/
    """

    optimizer: Optional[PolicyOptimizer]
    workers: Optional[WorkerSet]
    # Handle all config merging in RaylabOptions
    options: RaylabOptions = RaylabOptions()

    @overrides(RLlibTrainer)
    def train(self):
        result = {}

        # Run evaluation once before any optimization is done
        if self._iteration == 0 and self.config["evaluation_interval"]:
            result.update(self._evaluate())

        result.update(super().train())

        # Update global_vars after training so that they're saved if checkpointing
        self.global_vars["timestep"] = self.metrics.num_steps_sampled

        if self.config["wandb"]["enabled"]:
            self._wandb_log_result(result)
        return result

    @staticmethod
    def _wandb_log_result(result: dict):
        # Avoid logging the config every iteration
        # Only log Jsonable objects
        filtered = {
            k: v for k, v in result.items() if k != "config" and isinstance(v, Jsonable)
        }
        wandb.log(filtered)

    @property
    @overrides(RLlibTrainer)
    def _default_config(self):
        return self.options.defaults

    @overrides(RLlibTrainer)
    def _setup(self, config: dict):
        if not self.options.all_options_set:
            raise RuntimeError(
                f"{type(self).__name__} still has configs to be set."
                " Did you call `trainer.configure` as the last decorator?"
            )

        self.config = config = self.options.merge_defaults_with(config)

        self.env_creator = compat.make_env_creator(self._env_id, config)

        compat.check_and_resolve_framework_settings(config)

        RLlibTrainer._validate_config(config)

        self.callbacks = compat.validate_callbacks(config)

        compat.set_rllib_log_level(config)

        self._init(config, self.env_creator)

        # Evaluation setup.
        if config.get("evaluation_interval"):
            evaluation_config = compat.setup_evaluation_config(config)
            self.evaluation_workers = self._make_workers(
                self.env_creator,
                self._policy,
                evaluation_config,
                num_workers=config["evaluation_num_workers"],
            )

        self._setup_optimizer_placeholder()

        if self.config["compile_policy"]:
            self._optimize_policy_backend()

        if self.config["wandb"]["enabled"]:
            self._setup_wandb()

    def _setup_optimizer_placeholder(self):
        # Always have a PolicyOptimizer if possible to collect metrics
        if not hasattr(self, "optimizer"):
            if hasattr(self, "workers"):
                self.optimizer = PolicyOptimizer(self.workers)
            else:
                warnings.warn(
                    "No worker set initialized; episodes summary will be unavailable."
                )
        # Always have a WorkerSet if possible to get workers and policy
        if hasattr(self, "optimizer") and not hasattr(self, "workers"):
            self.workers = self.optimizer.workers

    def _optimize_policy_backend(self):
        if not hasattr(self, "workers"):
            raise RuntimeError(
                f"{type(self).__name__} has no worker set. "
                "Cannot access policies for compilation."
            )
        self.workers.foreach_policy(lambda p, _: p.compile())

    def _setup_wandb(self):
        assert wandb is not None, "Unable to import wandb, did you install it via pip?"

        wandb_kwargs = dict(
            name=self._name,
            config_exclude_keys={"wandb", "callbacks"},
            config=self.config,
            # Allow calling init twice if creating more than one trainer in the
            # same process
            reinit=True,
        )
        special_keys = {"enabled", "file_paths", "save_checkpoints"}
        wandb_kwargs.update(
            {k: v for k, v in self.config["wandb"].items() if k not in special_keys}
        )
        wandb.init(**wandb_kwargs)

        file_paths = self.config["wandb"]["file_paths"]
        for path in file_paths:
            wandb.save(path)

    def __getattr__(self, attr):
        if attr == "metrics":
            return self.optimizer

        raise AttributeError(f"{type(self).__name__} has no '{attr}' attribute")

    @classmethod
    @overrides(RLlibTrainer)
    def default_resource_request(cls, config: dict) -> compat.Resources:
        return compat.default_resource_request(cls, config)

    @overrides(Trainable)
    def save(self, checkpoint_dir=None):
        checkpoint_path = super().save(checkpoint_dir)

        config = self.config
        if config["wandb"]["enabled"] and config["wandb"]["save_checkpoints"]:
            wandb.save(checkpoint_path)

        return checkpoint_path

    @overrides(RLlibTrainer)
    def __getstate__(self):
        state = super().__getstate__()
        state["global_vars"] = self.global_vars
        return state

    @overrides(RLlibTrainer)
    def __setstate__(self, state):
        self.global_vars = state["global_vars"]
        if hasattr(self, "workers"):
            self.workers.foreach_worker(lambda w: w.set_global_vars(self.global_vars))

        super().__setstate__(state)

    @overrides(RLlibTrainer)
    def _stop(self):
        super()._stop()
        if self.config["wandb"]["enabled"] and wandb:
            wandb.join()



"""Losses for computing policy gradients."""
from typing import Optional
from typing import Tuple

import torch
from ray.rllib import SampleBatch
from torch import Tensor

from raylab.policy.modules.actor.policy.deterministic import DeterministicPolicy
from raylab.policy.modules.actor.policy.stochastic import StochasticPolicy
from raylab.policy.modules.critic.q_value import QValueEnsemble
from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict

from .abstract import Loss
from .utils import action_dpg


class DeterministicPolicyGradient(Loss):
    """Loss function for Deterministic Policy Gradient.

    Args:
        actor: deterministic policy
        critics: callables for action-values
    """

    batch_keys: Tuple[str] = (SampleBatch.CUR_OBS,)

    def __init__(self, actor: DeterministicPolicy, critics: QValueEnsemble):
        self.actor = actor
        self.critics = critics

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        obs = batch[SampleBatch.CUR_OBS]

        values = self.state_value(obs)
        loss = -torch.mean(values)

        stats = {"loss(actor)": loss.item()}
        return loss, stats

    def state_value(self, obs: Tensor) -> Tensor:
        """Compute the state value by combining policy and action-value function."""
        actions = self.actor(obs)
        return self.critics(obs, actions).min(dim=-1)[0]


class ReparameterizedSoftPG(Loss):
    """Loss function for Soft Policy Iteration with reparameterized actor.

    Args:
        actor: stochastic reparameterized policy
        critics: callables for action-values

    Attributes:
        alpha: entropy coefficient schedule
    """

    batch_keys: Tuple[str] = (SampleBatch.CUR_OBS,)
    alpha: float = 0.05

    def __init__(
        self, actor: StochasticPolicy, critics: QValueEnsemble,
    ):
        self.actor = actor
        self.critics = critics

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        obs = batch[SampleBatch.CUR_OBS]

        action_values, entropy = self.action_value_plus_entropy(obs)
        loss = -torch.mean(action_values + self.alpha * entropy)

        stats = {"loss(actor)": loss.item(), "entropy": entropy.mean().item()}
        return loss, stats

    def action_value_plus_entropy(self, obs: Tensor) -> Tuple[Tensor, Tensor]:
        """
        Compute the action-value and a single sample estimate of the policy's entropy.
        """
        actions, logp = self.actor.rsample(obs)
        action_values = self.critics(obs, actions).min(dim=-1)[0]
        return action_values, -logp


class ActionDPG(Loss):
    """Deterministic Policy Gradient by separating action and Q-value grads.

    Args:
        actor: deterministic policy
        critics: Q-value functions

    Attributes:
        dqda_clipping: Optional value by which to clip the action gradients
        clip_norm: Whether to clip action grads by norm or value
    """

    batch_keys: Tuple[str] = (SampleBatch.CUR_OBS,)
    dqda_clipping: Optional[float] = None
    clip_norm: bool = True

    def __init__(self, actor: DeterministicPolicy, critics: QValueEnsemble):
        self.actor = actor
        self.critics = critics

    def compile(self):
        self.actor = torch.jit.script(self.actor)
        self.critics = torch.jit.script(self.critics)

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        obs = batch[SampleBatch.CUR_OBS]
        a_max = self.actor(obs)
        q_max = self.critics(obs, a_max).min(dim=-1)[0]

        loss, dqda_norm = action_dpg(q_max, a_max, self.dqda_clipping, self.clip_norm)
        loss = loss.mean()
        return loss, {"loss(actor)": loss.item(), "dqda_norm": dqda_norm.mean().item()}



"""
MIT License

Copyright (c) 2019 Conor Durkan, Artur Bekasov, Iain Murray, George Papamakarios

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Slightly modified from:
https://github.com/bayesiains/nsf/blob/master/nde/transforms/splines/rational_quadratic.py
"""
# pylint:disable=missing-docstring,too-many-arguments,too-many-locals
import math

import torch
from torch.nn import functional as F


DEFAULT_MIN_BIN_WIDTH = 1e-3
DEFAULT_MIN_BIN_HEIGHT = 1e-3
DEFAULT_MIN_DERIVATIVE = 1e-3


def searchsorted(bin_locations, inputs, eps: float = 1e-6):
    bin_locations[..., -1] += eps
    return torch.sum(inputs[..., None] >= bin_locations, dim=-1) - 1


def unconstrained_rational_quadratic_spline(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    unnormalized_derivatives,
    inverse: bool = False,
    tail_bound: float = 1.0,
    min_bin_width: float = DEFAULT_MIN_BIN_WIDTH,
    min_bin_height: float = DEFAULT_MIN_BIN_HEIGHT,
    min_derivative: float = DEFAULT_MIN_DERIVATIVE,
):
    inside_interval_mask = (inputs >= -tail_bound) & (inputs <= tail_bound)
    outside_interval_mask = ~inside_interval_mask

    outputs = torch.zeros_like(inputs)
    logabsdet = torch.zeros_like(inputs)

    # Always use linear tails
    unnormalized_derivatives = F.pad(unnormalized_derivatives, pad=(1, 1))
    constant = math.log(math.exp(1 - min_derivative) - 1)
    unnormalized_derivatives[..., 0] = constant
    unnormalized_derivatives[..., -1] = constant

    outputs[outside_interval_mask] = inputs[outside_interval_mask]
    logabsdet[outside_interval_mask] = torch.zeros_like(inputs[outside_interval_mask])

    spline_out = rational_quadratic_spline(
        inputs=inputs[inside_interval_mask],
        unnormalized_widths=unnormalized_widths[inside_interval_mask, :],
        unnormalized_heights=unnormalized_heights[inside_interval_mask, :],
        unnormalized_derivatives=unnormalized_derivatives[inside_interval_mask, :],
        inverse=inverse,
        left=-tail_bound,
        right=tail_bound,
        bottom=-tail_bound,
        top=tail_bound,
        min_bin_width=min_bin_width,
        min_bin_height=min_bin_height,
        min_derivative=min_derivative,
    )
    outputs[inside_interval_mask], logabsdet[inside_interval_mask] = spline_out
    return outputs, logabsdet


def rational_quadratic_spline(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    unnormalized_derivatives,
    inverse: bool = False,
    left: float = 0.0,
    right: float = 1.0,
    bottom: float = 0.0,
    top: float = 1.0,
    min_bin_width: float = DEFAULT_MIN_BIN_WIDTH,
    min_bin_height: float = DEFAULT_MIN_BIN_HEIGHT,
    min_derivative: float = DEFAULT_MIN_DERIVATIVE,
):
    # if torch.min(inputs) < left or torch.max(inputs) > right:
    #     raise transforms.InputOutsideDomain()

    num_bins = unnormalized_widths.shape[-1]

    if min_bin_width * num_bins > 1.0:
        raise ValueError("Minimal bin width too large for the number of bins")
    if min_bin_height * num_bins > 1.0:
        raise ValueError("Minimal bin height too large for the number of bins")

    widths = F.softmax(unnormalized_widths, dim=-1)
    widths = min_bin_width + (1 - min_bin_width * num_bins) * widths
    cumwidths = torch.cumsum(widths, dim=-1)
    cumwidths = F.pad(cumwidths, pad=(1, 0), mode="constant", value=0.0)
    cumwidths = (right - left) * cumwidths + left
    cumwidths[..., 0] = left
    cumwidths[..., -1] = right
    widths = cumwidths[..., 1:] - cumwidths[..., :-1]

    derivatives = min_derivative + F.softplus(unnormalized_derivatives)

    heights = F.softmax(unnormalized_heights, dim=-1)
    heights = min_bin_height + (1 - min_bin_height * num_bins) * heights
    cumheights = torch.cumsum(heights, dim=-1)
    cumheights = F.pad(cumheights, pad=(1, 0), mode="constant", value=0.0)
    cumheights = (top - bottom) * cumheights + bottom
    cumheights[..., 0] = bottom
    cumheights[..., -1] = top
    heights = cumheights[..., 1:] - cumheights[..., :-1]

    if inverse:
        bin_idx = searchsorted(cumheights, inputs)[..., None]
    else:
        bin_idx = searchsorted(cumwidths, inputs)[..., None]

    input_cumwidths = cumwidths.gather(-1, bin_idx)[..., 0]
    input_bin_widths = widths.gather(-1, bin_idx)[..., 0]

    input_cumheights = cumheights.gather(-1, bin_idx)[..., 0]
    delta = heights / widths
    input_delta = delta.gather(-1, bin_idx)[..., 0]

    input_derivatives = derivatives.gather(-1, bin_idx)[..., 0]
    input_derivatives_plus_one = derivatives[..., 1:].gather(-1, bin_idx)[..., 0]

    input_heights = heights.gather(-1, bin_idx)[..., 0]

    if inverse:
        return rational_quadratic_inverse(
            inputs,
            input_cumheights,
            input_derivatives,
            input_derivatives_plus_one,
            input_delta,
            input_heights,
            input_bin_widths,
            input_cumwidths,
        )

    return rational_quadratic_forward(
        inputs,
        input_cumheights,
        input_derivatives,
        input_derivatives_plus_one,
        input_delta,
        input_heights,
        input_bin_widths,
        input_cumwidths,
    )


def rational_quadratic_forward(
    inputs,
    input_cumheights,
    input_derivatives,
    input_derivatives_plus_one,
    input_delta,
    input_heights,
    input_bin_widths,
    input_cumwidths,
):
    theta = (inputs - input_cumwidths) / input_bin_widths
    theta_one_minus_theta = theta * (1 - theta)

    numerator = input_heights * (
        input_delta * theta.pow(2) + input_derivatives * theta_one_minus_theta
    )
    denominator = input_delta + (
        (input_derivatives + input_derivatives_plus_one - 2 * input_delta)
        * theta_one_minus_theta
    )
    outputs = input_cumheights + numerator / denominator

    derivative_numerator = input_delta.pow(2) * (
        input_derivatives_plus_one * theta.pow(2)
        + 2 * input_delta * theta_one_minus_theta
        + input_derivatives * (1 - theta).pow(2)
    )
    logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)

    return outputs, logabsdet


def rational_quadratic_inverse(
    inputs,
    input_cumheights,
    input_derivatives,
    input_derivatives_plus_one,
    input_delta,
    input_heights,
    input_bin_widths,
    input_cumwidths,
):
    # pylint:disable=invalid-name
    a = (inputs - input_cumheights) * (
        input_derivatives + input_derivatives_plus_one - 2 * input_delta
    ) + input_heights * (input_delta - input_derivatives)
    b = input_heights * input_derivatives - (inputs - input_cumheights) * (
        input_derivatives + input_derivatives_plus_one - 2 * input_delta
    )
    c = -input_delta * (inputs - input_cumheights)

    discriminant = b.pow(2) - 4 * a * c
    assert (discriminant >= 0).all()

    root = (2 * c) / (-b - torch.sqrt(discriminant))
    outputs = root * input_bin_widths + input_cumwidths

    theta_one_minus_theta = root * (1 - root)
    denominator = input_delta + (
        (input_derivatives + input_derivatives_plus_one - 2 * input_delta)
        * theta_one_minus_theta
    )
    derivative_numerator = input_delta.pow(2) * (
        input_derivatives_plus_one * root.pow(2)
        + 2 * input_delta * theta_one_minus_theta
        + input_derivatives * (1 - root).pow(2)
    )
    logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)

    return outputs, -logabsdet



"""SVG model architecture."""
import torch
import torch.nn as nn
from ray.rllib.utils import override

import raylab.pytorch.nn as nnx
import raylab.pytorch.nn.distributions as ptd
from raylab.pytorch.nn.init import initialize_
from raylab.utils.dictionaries import deep_merge

from .stochastic_model_mixin import StochasticModel


BASE_CONFIG = {
    "residual": True,
    "input_dependent_scale": False,
    "encoder": {"units": (40, 40), "activation": "Tanh", "delay_action": True},
    "initializer_options": {"name": "xavier_uniform"},
}


class SVGModelMixin:
    # pylint:disable=missing-docstring,too-few-public-methods
    @staticmethod
    def _make_model(obs_space, action_space, config):
        config = deep_merge(BASE_CONFIG, config.get("model", {}), False, ["encoder"])

        params_module = SVGDynamicsParams(obs_space, action_space, config)
        dist_module = ptd.Independent(ptd.Normal(), reinterpreted_batch_ndims=1)

        model = StochasticModel.assemble(params_module, dist_module, config)
        return {"model": model}


class SVGDynamicsParams(nn.Module):
    """
    Neural network module mapping inputs to distribution parameters through parallel
    subnetworks for each output dimension.
    """

    def __init__(self, obs_space, action_space, config):
        super().__init__()
        obs_size, act_size = obs_space.shape[0], action_space.shape[0]

        def make_encoder():
            mlp = nnx.StateActionEncoder(obs_size, act_size, **config["encoder"])
            mlp.apply(
                initialize_(
                    activation=config["encoder"].get("activation"),
                    **config["initializer_options"],
                )
            )
            return mlp

        self.logits = nn.ModuleList([make_encoder() for _ in range(obs_space.shape[0])])

        def make_param(in_features):
            kwargs = dict(event_size=1, input_dependent_scale=False)
            return nnx.NormalParams(in_features, **kwargs)

        self.params = nn.ModuleList([make_param(m.out_features) for m in self.logits])

    @override(nn.Module)
    def forward(self, obs, act):  # pylint:disable=arguments-differ
        params = [p(l(obs, act)) for p, l in zip(self.params, self.logits)]
        loc = torch.cat([d["loc"] for d in params], dim=-1)
        scale = torch.cat([d["scale"] for d in params], dim=-1)
        return {"loc": loc, "scale": scale}



"""Utilities for manipulating python dictionaries and general mappings."""
import copy

from ray.rllib.utils import deep_update


def deep_merge(dict1, dict2, *args, **kwargs):
    """Deep copy original dict and pass it to RLlib's deep_update."""
    clone = copy.deepcopy(dict1)
    return deep_update(clone, dict2, *args, **kwargs)


def get_keys(mapping, *keys):
    """Return the values corresponding to the given keys, in order."""
    return (mapping[k] for k in keys)


def all_except(mapping, *exclude):
    """Return a new mapping with all keys except `exclude`.

    Keys must be hashable to be used with `set`.
    """
    exclude = set(exclude)
    return {k: v for k, v in mapping.items() if k not in exclude}



# pylint:disable=missing-docstring
# pylint:enable=missing-docstring
from typing import Dict
from typing import List

import torch
import torch.nn as nn
import torch.nn.functional as F
from ray.rllib.utils import override

from .abstract import ConditionalDistribution


class Categorical(ConditionalDistribution):
    r"""
    Creates a categorical distribution parameterized by `logits`.

    .. note::
        It is equivalent to the distribution that :func:`torch.multinomial`
        samples from.

    Samples are integers from
        :math:`\{0, \ldots, K-1\}` where `K` is ``logits.size(-1)``.

    Example::

        >>> params = {"logits": torch.tensor([0.25, 0.25, 0.25, 0.25])}
        >>> m = Categorical()
        >>> m.sample(params)  # equal probability of 0, 1, 2, 3
        tensor(3)
    """

    @override(nn.Module)
    def forward(self, inputs):  # pylint:disable=arguments-differ
        return {"logits": inputs - inputs.logsumexp(dim=-1, keepdim=True)}

    @override(ConditionalDistribution)
    @torch.jit.export
    def sample(self, params: Dict[str, torch.Tensor], sample_shape: List[int] = ()):
        logits = self._unpack_params(params)
        params_shape = sample_shape + logits.shape
        sample_shape = sample_shape + logits.shape[:-1]
        probs = F.softmax(logits, dim=-1).expand(params_shape)
        probs_2d = probs.reshape(-1, logits.shape[-1])
        sample_2d = torch.multinomial(probs_2d, 1, True)
        out = sample_2d.reshape(sample_shape)
        return out, self.log_prob(out, params)

    @override(ConditionalDistribution)
    @torch.jit.export
    def log_prob(self, value, params: Dict[str, torch.Tensor]):
        logits = self._unpack_params(params)
        value = value.long().unsqueeze(-1)
        value, log_pmf = torch.broadcast_tensors(value, logits)
        value = value[..., :1]
        return log_pmf.gather(-1, value).squeeze(-1)

    @override(ConditionalDistribution)
    @torch.jit.export
    def entropy(self, params: Dict[str, torch.Tensor]):
        logits = self._unpack_params(params)
        probs = F.softmax(logits, dim=-1)
        p_log_p = logits * probs
        return -p_log_p.sum(-1)

    @override(ConditionalDistribution)
    @torch.jit.export
    def deterministic(self, params: Dict[str, torch.Tensor]):
        logits = self._unpack_params(params)
        sample = torch.argmax(logits, dim=-1)
        return sample, self.log_prob(sample, params)

    def _unpack_params(self, params: Dict[str, torch.Tensor]):
        # pylint:disable=no-self-use
        return params["logits"]



"""NN modules for stochastic dynamics estimation."""
from typing import List

import torch
import torch.nn as nn
from gym.spaces import Box

import raylab.pytorch.nn as nnx
import raylab.pytorch.nn.distributions as ptd
from raylab.policy.modules.networks.mlp import StateActionMLP


class StochasticModel(nn.Module):
    """Represents a stochastic model as a conditional distribution module."""

    def __init__(
        self, params_module: nn.Module, dist_module: ptd.ConditionalDistribution
    ):
        super().__init__()
        self.params = params_module
        self.dist = dist_module

    def forward(self, obs, action):  # pylint:disable=arguments-differ
        return self.params(obs, action)

    @torch.jit.export
    def sample(self, obs, action, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped sample or sample_shape shaped batch of
        samples if the distribution parameters are batched. Returns a (sample, log_prob)
        pair.
        """
        params = self(obs, action)
        return self.dist.sample(params, sample_shape)

    @torch.jit.export
    def rsample(self, obs, action, sample_shape: List[int] = ()):
        """
        Generates a sample_shape shaped reparameterized sample or sample_shape
        shaped batch of reparameterized samples if the distribution parameters
        are batched. Returns a (rsample, log_prob) pair.
        """
        params = self(obs, action)
        return self.dist.rsample(params, sample_shape)

    @torch.jit.export
    def log_prob(self, obs, action, next_obs):
        """
        Returns the log probability density/mass function evaluated at `next_obs`.
        """
        params = self(obs, action)
        return self.dist.log_prob(next_obs, params)

    @torch.jit.export
    def cdf(self, obs, action, next_obs):
        """Returns the cumulative density/mass function evaluated at `next_obs`."""
        params = self(obs, action)
        return self.dist.cdf(next_obs, params)

    @torch.jit.export
    def icdf(self, obs, action, prob):
        """Returns the inverse cumulative density/mass function evaluated at `prob`."""
        params = self(obs, action)
        return self.dist.icdf(prob, params)

    @torch.jit.export
    def entropy(self, obs, action):
        """Returns entropy of distribution."""
        params = self(obs, action)
        return self.dist.entropy(params)

    @torch.jit.export
    def perplexity(self, obs, action):
        """Returns perplexity of distribution."""
        params = self(obs, action)
        return self.dist.perplexity(params)

    @torch.jit.export
    def reproduce(self, obs, action, next_obs):
        """Produce a reparametrized sample with the same value as `next_obs`."""
        params = self(obs, action)
        return self.dist.reproduce(next_obs, params)


class ResidualMixin:
    """Overrides StochasticModel interface to model state transition residuals."""

    # pylint:disable=missing-function-docstring,not-callable

    @torch.jit.export
    def sample(self, obs, action, sample_shape: List[int] = ()):
        params = self(obs, action)
        res, log_prob = self.dist.sample(params, sample_shape)
        return obs + res, log_prob

    @torch.jit.export
    def rsample(self, obs, action, sample_shape: List[int] = ()):
        params = self(obs, action)
        res, log_prob = self.dist.rsample(params, sample_shape)
        return obs + res, log_prob

    @torch.jit.export
    def log_prob(self, obs, action, next_obs):
        params = self(obs, action)
        return self.dist.log_prob(next_obs - obs, params)

    @torch.jit.export
    def cdf(self, obs, action, next_obs):
        params = self(obs, action)
        return self.dist.cdf(next_obs - obs, params)

    @torch.jit.export
    def icdf(self, obs, action, prob):
        params = self(obs, action)
        residual = self.dist.icdf(prob, params)
        return obs + residual

    @torch.jit.export
    def reproduce(self, obs, action, next_obs):
        params = self(obs, action)
        sample_, log_prob_ = self.dist.reproduce(next_obs - obs, params)
        return obs + sample_, log_prob_


class DynamicsParams(nn.Module):
    """Neural network mapping state-action pairs to distribution parameters.

    Args:
        encoder: Module mapping state-action pairs to 1D features
        params: Module mapping 1D features to distribution parameters
    """

    def __init__(self, encoder: nn.Module, params: nn.Module):
        super().__init__()
        self.encoder = encoder
        self.params = params

    def forward(self, obs, actions):  # pylint:disable=arguments-differ
        return self.params(self.encoder(obs, actions))


MLPSpec = StateActionMLP.spec_cls


class MLPModel(StochasticModel):
    """Stochastic model with multilayer perceptron state-action encoder."""

    spec_cls = MLPSpec

    def __init__(
        self,
        obs_space: Box,
        action_space: Box,
        spec: MLPSpec,
        input_dependent_scale: bool,
    ):
        encoder = StateActionMLP(obs_space, action_space, spec)
        params = nnx.NormalParams(
            encoder.out_features,
            obs_space.shape[0],
            input_dependent_scale=input_dependent_scale,
        )
        params = DynamicsParams(encoder, params)
        dist = ptd.Independent(ptd.Normal(), reinterpreted_batch_ndims=1)
        super().__init__(params, dist)
        self.encoder = encoder

    def initialize_parameters(self, initializer_spec: dict):
        """Initialize all encoder parameters.

        Args:
            initializer_spec: Dictionary with mandatory `name` key corresponding
                to the initializer function name in `torch.nn.init` and optional
                keyword arguments.
        """
        self.encoder.initialize_parameters(initializer_spec)


class ResidualMLPModel(ResidualMixin, MLPModel):
    """Residual stochastic multilayer perceptron model."""



"""Flow implementations.

Each flow is invertible so it can be computed (if training) and inverted
(if evaluating). Each flow also outputs its log of the Jacobian determinant.

References:
    Papamakarios, George, et al. "Normalizing Flows for Probabilistic Modeling
    and Inference." arXiv preprint arXiv:1912.02762 (2019).
"""

from . import masks
from . import utils
from .abstract import CompositeTransform
from .abstract import ConditionalTransform
from .abstract import InverseTransform
from .abstract import Transform
from .affine_constant import ActNorm
from .affine_constant import AffineConstantFlow
from .coupling import AdditiveCouplingTransform
from .coupling import AffineCouplingTransform
from .coupling import CouplingTransform
from .coupling import PiecewiseRQSCouplingTransform
from .maf import IAF
from .maf import MAF
from .nonlinearities import AffineTransform
from .nonlinearities import SigmoidSquashTransform
from .nonlinearities import SigmoidTransform
from .nonlinearities import TanhSquashTransform
from .nonlinearities import TanhTransform


__all__ = [
    "ActNorm",
    "AdditiveCouplingTransform",
    "AffineConstantFlow",
    "AffineCouplingTransform",
    "AffineTransform",
    "ConditionalTransform",
    "CouplingTransform",
    "IAF",
    "InverseTransform",
    "MAF",
    "masks",
    "PiecewiseRQSCouplingTransform",
    "SigmoidTransform",
    "SigmoidSquashTransform",
    "TanhTransform",
    "TanhSquashTransform",
    "Transform",
]



"""Utilities for visualization."""
import bokeh
import numpy as np
import pandas as pd
from bokeh.models import BoxZoomTool
from bokeh.models import ColumnDataSource
from bokeh.models import HelpTool
from bokeh.models import HoverTool
from bokeh.models import PanTool
from bokeh.models import ResetTool
from bokeh.models import SaveTool
from bokeh.models import WheelZoomTool
from bokeh.plotting import figure


def time_series(x_key, y_key, groups, labels, config):
    """Plot time series with error bands per group."""
    # pylint:disable=too-many-function-args
    kwargs = dict(y_axis_type="log") if config["log_scale"] else {}
    pic = figure(title="Plot", **kwargs)
    pic.xaxis.axis_label = x_key
    pic.yaxis.axis_label = y_key

    pic.tools = [
        PanTool(),
        BoxZoomTool(),
        WheelZoomTool(dimensions="height"),
        WheelZoomTool(dimensions="width"),
        SaveTool(),
        ResetTool(),
        HelpTool(),
    ]
    pic.add_tools()
    if config["individual"]:
        pic.add_tools(HoverTool(tooltips=[("y", "@y"), ("x", "@x{a}"), ("id", "@id")]))
    else:
        pic.add_tools(HoverTool(tooltips=[("y", "@y_mean"), ("x", "@x{a}")]))

    for label, group, color in zip(labels, groups, bokeh.palettes.cividis(len(labels))):
        data = group.extract()
        progresses = [d.progress for d in data if y_key in d.progress.columns]
        if not progresses:
            continue

        x_all, all_ys = filter_and_interpolate(x_key, y_key, progresses)

        if config["individual"]:
            plot_individual(pic, x_all, all_ys, data, label, color)
        else:
            plot_mean_dispersion(
                pic,
                x_all,
                all_ys,
                label,
                color,
                standard_error=config["standard_error"],
            )

    pic.legend.location = "bottom_left"
    pic.legend.click_policy = "hide"
    return pic


def filter_and_interpolate(x_key, y_key, progresses):
    # pylint:disable=missing-function-docstring
    # Filter NaN values from plots
    masks = [~np.isnan(p[y_key]) for p in progresses]
    xs_ = [p[x_key][m] for m, p in zip(masks, progresses)]
    ys_ = [p[y_key][m] for m, p in zip(masks, progresses)]
    x_all = np.unique(np.sort(np.concatenate(xs_)))
    all_ys = [
        np.interp(x_all, x, y, left=np.nan, right=np.nan) for x, y in zip(xs_, ys_)
    ]
    return x_all, all_ys


def plot_individual(pic, x_all, all_ys, data, label, color):
    # pylint:disable=missing-function-docstring,too-many-arguments
    for datum, y_i in zip(data, all_ys):
        identifier = str(datum.params["id"])
        dataframe = pd.DataFrame({"x": x_all, "y": y_i, "id": identifier})
        source = ColumnDataSource(data=dataframe)
        pic.line(
            x="x", y="y", source=source, legend_label=label, color=color,
        )


def plot_mean_dispersion(pic, x_all, all_ys, label, color, standard_error=False):
    # pylint:disable=missing-function-docstring,too-many-arguments
    y_mean = np.nanmean(all_ys, axis=0)
    dispersion = np.nanstd(all_ys, axis=0)
    if standard_error:
        dispersion /= np.sqrt(np.sum(1 - np.isnan(y_mean), axis=0))
    dataframe = pd.DataFrame(
        {
            "x": x_all,
            "y_mean": y_mean,
            "y_low": y_mean - dispersion,
            "y_high": y_mean + dispersion,
            "label": label,
        }
    )
    source = ColumnDataSource(data=dataframe)
    pic.line(x="x", y="y_mean", source=source, legend_label=label, color=color)
    pic.varea(
        x="x",
        y1="y_low",
        y2="y_high",
        source=source,
        fill_alpha=0.25,
        legend_label=label,
        color=color,
    )



"""SVG(1) policy class using PyTorch."""
import warnings

import torch
import torch.nn as nn
from ray.rllib import SampleBatch
from ray.rllib.utils import override

from raylab.agents.svg import SVGTorchPolicy
from raylab.policy import AdaptiveKLCoeffMixin
from raylab.policy import EnvFnMixin
from raylab.policy import TorchPolicy
from raylab.policy.losses import OneStepSVG
from raylab.pytorch.optim import get_optimizer_class


class SVGOneTorchPolicy(AdaptiveKLCoeffMixin, SVGTorchPolicy):
    """Stochastic Value Gradients policy for off-policy learning."""

    # pylint:disable=abstract-method

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loss_actor = OneStepSVG(
            self.module.model.reproduce,
            self.module.actor.reproduce,
            self.module.critic,
        )
        self.loss_actor.gamma = self.config["gamma"]

    @property
    @override(SVGTorchPolicy)
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.svg.one import SVGOneTrainer

        return SVGOneTrainer.options

    @override(TorchPolicy)
    def compile(self):
        warnings.warn(f"{type(self).__name__} is incompatible with TorchScript")

    def update_old_policy(self):
        """Copy params of current policy into old one for future KL computation."""
        self.module.old_actor.load_state_dict(self.module.actor.state_dict())

    @override(EnvFnMixin)
    def _set_reward_hook(self):
        self.loss_actor.set_reward_fn(self.reward_fn)

    @override(SVGTorchPolicy)
    def _make_module(self, obs_space, action_space, config):
        config["module"]["replay_kl"] = config["replay_kl"]
        return super()._make_module(obs_space, action_space, config)

    @override(SVGTorchPolicy)
    def _make_optimizers(self):
        """PyTorch optimizer to use."""
        optimizers = super()._make_optimizers()
        cls = get_optimizer_class(self.config["torch_optimizer"]["type"], wrap=True)
        options = self.config["torch_optimizer"]
        modules = {
            "model": self.module.model,
            "actor": self.module.actor,
            "critic": self.module.critic,
        }
        param_groups = [
            dict(params=mod.parameters(), **options[name])
            for name, mod in modules.items()
        ]
        optimizers["all"] = cls(param_groups)
        return optimizers

    @override(SVGTorchPolicy)
    def learn_on_batch(self, samples):
        batch_tensors = self.lazy_tensor_dict(samples)
        batch_tensors, info = self.add_truncated_importance_sampling_ratios(
            batch_tensors
        )

        with self.optimizers.optimize("all"):
            model_value_loss, stats = self.compute_joint_model_value_loss(batch_tensors)
            info.update(stats)
            model_value_loss.backward()

            self.module.model.requires_grad_(False)
            self.module.critic.requires_grad_(False)

            svg_loss, stats = self.loss_actor(batch_tensors)
            info.update(stats)
            kl_loss = self.curr_kl_coeff * self._avg_kl_divergence(batch_tensors)
            (svg_loss + kl_loss).backward()

            self.module.model.requires_grad_(True)
            self.module.critic.requires_grad_(True)

        info.update(self.extra_grad_info(batch_tensors))
        info.update(self.update_kl_coeff(samples))
        self._update_polyak()
        return info

    @torch.no_grad()
    @override(AdaptiveKLCoeffMixin)
    def _kl_divergence(self, sample_batch):
        batch_tensors = self.lazy_tensor_dict(sample_batch)
        return self._avg_kl_divergence(batch_tensors).item()

    def _avg_kl_divergence(self, batch_tensors):
        if self.config["replay_kl"]:
            logp = self.module.actor.log_prob(
                batch_tensors[SampleBatch.CUR_OBS], batch_tensors[SampleBatch.ACTIONS]
            )
            return torch.mean(batch_tensors[SampleBatch.ACTION_LOGP] - logp)

        old_act, old_logp = self.module.old_actor.rsample(
            batch_tensors[SampleBatch.CUR_OBS]
        )
        logp = self.module.actor.log_prob(batch_tensors[SampleBatch.CUR_OBS], old_act)
        return torch.mean(old_logp - logp)

    @torch.no_grad()
    def extra_grad_info(self, batch_tensors):
        """Compute gradient norms and policy statistics."""
        grad_norms = {
            f"grad_norm({k})": nn.utils.clip_grad_norm_(
                self.module[k].parameters(), float("inf")
            )
            for k in "model actor critic".split()
        }
        policy_info = {
            "entropy": self.module.actor.log_prob(
                batch_tensors[SampleBatch.CUR_OBS], batch_tensors[SampleBatch.ACTIONS]
            )
            .mean()
            .neg()
            .item(),
            "curr_kl_coeff": self.curr_kl_coeff,
        }
        return {**grad_norms, **policy_info}



# pylint:disable=missing-docstring
import torch
import torch.nn as nn
from torch import Tensor


class TanhSquash(nn.Module):
    """Neural network module squashing vectors to specified range using Tanh."""

    def __init__(self, low: Tensor, high: Tensor):
        super().__init__()
        self.register_buffer("loc", (high + low) / 2)
        self.register_buffer("scale", (high - low) / 2)

    def forward(self, inputs: Tensor, reverse: bool = False) -> Tensor:
        # pylint:disable=arguments-differ
        if reverse:
            inputs = (inputs - self.loc) / self.scale
            to_log1 = torch.clamp(1 + inputs, min=1.1754943508222875e-38)
            to_log2 = torch.clamp(1 - inputs, min=1.1754943508222875e-38)
            return (torch.log(to_log1) - torch.log(to_log2)) / 2
        return self.loc + inputs.tanh() * self.scale



"""SVG Module with RealNVP density approximation for the policy."""

from .abstract import AbstractModelActorCritic
from .mixins import NormalizingFlowActorMixin
from .mixins import StateValueMixin
from .mixins import SVGModelMixin


# pylint:disable=abstract-method
class SVGRealNVPActor(
    SVGModelMixin, NormalizingFlowActorMixin, StateValueMixin, AbstractModelActorCritic
):
    """SVG Module with RealNVP density approximation for the policy."""



"""SOP policy class using PyTorch."""
import torch
import torch.nn as nn
from ray.rllib.utils import override

from raylab.policy import TorchPolicy
from raylab.policy.action_dist import WrapDeterministicPolicy
from raylab.policy.losses import ActionDPG
from raylab.policy.losses import ClippedDoubleQLearning
from raylab.policy.losses import DeterministicPolicyGradient
from raylab.pytorch.nn.utils import update_polyak
from raylab.pytorch.optim import build_optimizer


class SOPTorchPolicy(TorchPolicy):
    """Streamlined Off-Policy policy in PyTorch to use with RLlib."""

    # pylint:disable=abstract-method
    dist_class = WrapDeterministicPolicy

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._make_actor_loss()
        self.loss_critic = ClippedDoubleQLearning(
            self.module.critics, self.module.target_critics, self.module.target_actor,
        )
        self.loss_critic.gamma = self.config["gamma"]
        self._grad_step = 0

    @property
    @override(TorchPolicy)
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.sop import SOPTrainer

        return SOPTrainer.options

    def _make_actor_loss(self):
        if self.config["dpg_loss"] == "default":
            self.loss_actor = DeterministicPolicyGradient(
                self.module.actor, self.module.critics,
            )
        elif self.config["dpg_loss"] == "acme":
            self.loss_actor = ActionDPG(self.module.actor, self.module.critics)
            self.loss_actor.dqda_clipping = self.config["dqda_clipping"]
            self.loss_actor.clip_norm = self.config["clip_dqda_norm"]
        else:
            dpg_loss = self.config["dpg_loss"]
            raise ValueError(
                f"Invalid config for 'dpg_loss': {dpg_loss}."
                " Choose between 'default' and 'acme'"
            )

    @override(TorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["torch_optimizer"]
        components = "actor critics".split()

        mapping = {
            name: build_optimizer(getattr(self.module, name), config[name])
            for name in components
        }

        optimizers.update(mapping)
        return optimizers

    @override(TorchPolicy)
    def learn_on_batch(self, samples):
        batch_tensors = self.lazy_tensor_dict(samples)

        info = {}
        self._grad_step += 1
        info.update(self._update_critic(batch_tensors))
        if self._grad_step % self.config["policy_delay"] == 0:
            info.update(self._update_policy(batch_tensors))

        update_polyak(
            self.module.critics, self.module.target_critics, self.config["polyak"]
        )
        return info

    def _update_critic(self, batch_tensors):
        with self.optimizers.optimize("critics"):
            loss, info = self.loss_critic(batch_tensors)
            loss.backward()

        info.update(self.extra_grad_info("critics"))
        return info

    def _update_policy(self, batch_tensors):
        with self.optimizers.optimize("actor"):
            loss, info = self.loss_actor(batch_tensors)
            loss.backward()

        info.update(self.extra_grad_info("actor"))
        return info

    @torch.no_grad()
    def extra_grad_info(self, component):
        """Return statistics right after components are updated."""
        return {
            f"grad_norm({component})": nn.utils.clip_grad_norm_(
                getattr(self.module, component).parameters(), float("inf")
            ).item()
        }

    @override(TorchPolicy)
    def get_weights(self):
        weights = super().get_weights()
        weights["grad_step"] = self._grad_step
        return weights

    @override(TorchPolicy)
    def set_weights(self, weights):
        self._grad_step = weights["grad_step"]
        super().set_weights(weights)



"""
MIT License

Copyright (c) 2019 Conor Durkan, Artur Bekasov, Iain Murray, George Papamakarios

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Slightly modified from:
https://github.com/bayesiains/nsf/blob/master/nn/resnet.py
"""
from typing import Dict
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from raylab.pytorch.nn.utils import get_activation


class ResidualBlock(nn.Module):
    """A general-purpose residual block. Works only with 1-dim inputs."""

    def __init__(
        self,
        features,
        state_features=None,
        activation="ReLU",
        dropout_probability=0.0,
        use_batch_norm=False,
        zero_initialization=True,
    ):
        # pylint:disable=too-many-arguments
        super().__init__()
        activation = get_activation(activation)

        layers = []
        if use_batch_norm:
            layers += [nn.BatchNorm1d(features, eps=1e-3)]
        layers += [activation()]
        layers += [nn.Linear(features, features)]

        if use_batch_norm:
            layers += [nn.BatchNorm1d(features, eps=1e-3)]
        layers += [activation()]
        layers += [nn.Dropout(p=dropout_probability)]
        layers += [nn.Linear(features, features)]
        if zero_initialization:
            nn.init.uniform_(layers[-1].weight, -1e-3, 1e-3)
            nn.init.uniform_(layers[-1].bias, -1e-3, 1e-3)

        self.sequential = nn.Sequential(*layers)

        self.state_layer = None
        if state_features is not None:
            self.state_layer = nn.Linear(state_features, features)

    def forward(self, inputs, params: Optional[Dict[str, torch.Tensor]] = None):
        # pylint:disable=arguments-differ
        temps = inputs
        temps = self.sequential(temps)
        if self.state_layer is not None:
            if params is None:
                raise ValueError("Parameters required for stateful block.")
            temps = F.glu(
                torch.cat((temps, self.state_layer(params["state"])), dim=-1), dim=-1
            )
        return inputs + temps


class ResidualNet(nn.Module):
    """A general-purpose residual network. Works only with 1-dim inputs."""

    def __init__(
        self,
        in_features,
        out_features,
        hidden_features,
        state_features=None,
        num_blocks=2,
        activation="ReLU",
        dropout_probability=0.0,
        use_batch_norm=False,
    ):
        # pylint:disable=too-many-arguments
        super().__init__()
        self.hidden_features = hidden_features
        self.stateful = bool(state_features)
        if self.stateful:
            self.initial_layer = nn.Linear(
                in_features + state_features, hidden_features
            )
        else:
            self.initial_layer = nn.Linear(in_features, hidden_features)
        self.blocks = nn.ModuleList(
            [
                ResidualBlock(
                    features=hidden_features,
                    state_features=state_features,
                    activation=activation,
                    dropout_probability=dropout_probability,
                    use_batch_norm=use_batch_norm,
                )
                for _ in range(num_blocks)
            ]
        )
        self.final_layer = nn.Linear(hidden_features, out_features)

    def forward(self, inputs, params: Optional[Dict[str, torch.Tensor]] = None):
        # pylint:disable=arguments-differ
        if self.stateful:
            if params is None:
                raise ValueError("Parameters required for stateful resnet.")
            temps = self.initial_layer(torch.cat((inputs, params["state"]), dim=-1))
        else:
            temps = self.initial_layer(inputs)
        for block in self.blocks:
            temps = block(temps, params)
        outputs = self.final_layer(temps)
        return outputs



"""Policy for MAPO using PyTorch."""
from ray.rllib.utils import override

from raylab.agents.sac import SACTorchPolicy
from raylab.policy import EnvFnMixin
from raylab.policy import ModelTrainingMixin
from raylab.policy.losses import MAPO
from raylab.policy.losses import ModelEnsembleMLE
from raylab.policy.losses import SPAML
from raylab.pytorch.optim import build_optimizer


class MAPOTorchPolicy(ModelTrainingMixin, EnvFnMixin, SACTorchPolicy):
    """Model-Aware Policy Optimization policy in PyTorch to use with RLlib."""

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)
        self._setup_model_loss()

    @property
    def options(self):
        """Return the default configuration for MAPO."""
        # pylint:disable=cyclic-import
        from raylab.agents.mapo import MAPOTrainer

        return MAPOTrainer.options

    @property
    def model_training_loss(self):
        return self.loss_paml

    @property
    def model_warmup_loss(self):
        return self.loss_mle

    def compile(self):
        super().compile()
        self.loss_paml.compile()
        self.loss_mle.compile()
        self.loss_actor.compile()

    @override(ModelTrainingMixin)
    def optimize_model(self, *args, **kwargs):
        # pylint:disable=signature-differs
        self.loss_paml.alpha = self.module.alpha().item()
        return super().optimize_model(*args, **kwargs)

    def _setup_model_loss(self):
        self.loss_paml = SPAML(
            self.module.models, self.module.actor, self.module.critics
        )
        self.loss_paml.gamma = self.config["gamma"]
        self.loss_paml.grad_estimator = self.config["losses"]["grad_estimator"]
        self.loss_paml.lambda_ = self.config["losses"]["lambda"]

        self.loss_mle = ModelEnsembleMLE(self.module.models)

    @override(SACTorchPolicy)
    def _setup_actor_loss(self):
        self.loss_actor = MAPO(
            self.module.models, self.module.actor, self.module.critics
        )
        self.loss_actor.gamma = self.config["gamma"]
        self.loss_actor.model_samples = self.config["losses"]["model_samples"]
        self.loss_actor.grad_estimator = self.config["losses"]["grad_estimator"]

    @override(SACTorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["torch_optimizer"]["models"]
        optimizers["models"] = build_optimizer(self.module.models, config)
        return optimizers

    @override(EnvFnMixin)
    def _set_reward_hook(self):
        self.loss_paml.set_reward_fn(self.reward_fn)
        self.loss_actor.set_reward_fn(self.reward_fn)

    @override(EnvFnMixin)
    def _set_termination_hook(self):
        self.loss_paml.set_termination_fn(self.termination_fn)
        self.loss_actor.set_termination_fn(self.termination_fn)

    @override(EnvFnMixin)
    def _set_dynamics_hook(self):
        raise ValueError(
            """Model-Aware Policy Optimization shouldn't use ground-truth \
            dynamics. Refer to DAPO instead."""
        )



"""General module classes."""
import torch.nn as nn


class AbstractActorCritic(nn.ModuleDict):
    """Generic Actor-Critic architecture.

    Abstract module containing policy and value functions with no weight sharing.
    """

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        super().__init__()
        self.update(self._make_actor(obs_space, action_space, config))
        self.update(self._make_critic(obs_space, action_space, config))


class AbstractModelActorCritic(nn.ModuleDict):
    """Module containing env model, policy, and value functions.

    Generic model-based architecture with disjoint model, actor, and critic.
    """

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        super().__init__()
        self.update(self._make_model(obs_space, action_space, config))
        self.update(self._make_critic(obs_space, action_space, config))
        self.update(self._make_actor(obs_space, action_space, config))



"""Support for modules with action value functions as critics."""
import torch.nn as nn
from ray.rllib.utils import override

from raylab.pytorch.nn import StateActionEncoder
from raylab.pytorch.nn.init import initialize_
from raylab.utils.dictionaries import deep_merge


BASE_CONFIG = {
    "double_q": False,
    "initializer_options": {"name": "xavier_uniform"},
    "encoder": {"units": (32, 32), "activation": "ReLU", "delay_action": True},
}


class ActionValueMixin:
    """Adds constructor for modules with action value functions.

    Since it is common to use clipped double Q-Learning, critic is implemented as
    a ModuleList of action-value functions.
    """

    # pylint:disable=too-few-public-methods

    @staticmethod
    def _make_critic(obs_space, action_space, config):
        config = deep_merge(BASE_CONFIG, config.get("critic", {}), False, ["encoder"])
        obs_size, act_size = obs_space.shape[0], action_space.shape[0]

        def make_critic():
            return ActionValueFunction.from_scratch(
                obs_size,
                act_size,
                initializer_options=config["initializer_options"],
                **config["encoder"]
            )

        n_critics = 2 if config["double_q"] else 1
        critics = nn.ModuleList([make_critic() for _ in range(n_critics)])
        target_critics = nn.ModuleList([make_critic() for _ in range(n_critics)])
        target_critics.load_state_dict(critics.state_dict())
        return {"critics": critics, "target_critics": target_critics}


class ActionValueFunction(nn.Module):
    """Neural network module emulating a Q value function."""

    def __init__(self, logits_module, value_module):
        super().__init__()
        self.logits_module = logits_module
        self.value_module = value_module

    @override(nn.Module)
    def forward(self, obs, actions):  # pylint:disable=arguments-differ
        logits = self.logits_module(obs, actions)
        return self.value_module(logits)

    @classmethod
    def from_scratch(
        cls, obs_size, act_size, initializer_options=None, **logits_kwargs
    ):
        """Create an action value function with new logits and value modules."""
        logits_module = StateActionEncoder(obs_size, act_size, **logits_kwargs)
        initializer_options = initializer_options or {}
        logits_module.apply(
            initialize_(
                activation=logits_kwargs.get("activation"), **initializer_options
            )
        )
        value_module = nn.Linear(logits_module.out_features, 1)
        return cls(logits_module, value_module)



"""Network and configurations for modules with stochastic model ensembles."""
from typing import List
from typing import Tuple

import torch
import torch.nn as nn
from torch import Tensor

from .single import StochasticModel


class StochasticModelEnsemble(nn.ModuleList):
    """A static list of stochastic dynamics models.

    Args:
        models: List of StochasticModel modules
    """

    # pylint:disable=abstract-method

    def __init__(self, models: List[StochasticModel]):
        cls_name = type(self).__name__
        assert all(
            isinstance(m, StochasticModel) for m in models
        ), f"All modules in {cls_name} must be instances of StochasticModel."
        super().__init__(models)

    @torch.jit.export
    def sample(
        self, obs: Tensor, action: Tensor, sample_shape: List[int] = ()
    ) -> Tuple[Tensor, Tensor]:
        """Compute samples and likelihoods for each model in the ensemble.

        Splits inputs by the first dimension in `N` chunks, where `N` is the
        ensemble size. Then, applies each model in the ensemble to one of the
        slices.

        `O` is the observation shape and `A` is the action shape.

        Args:
            obs: Observation tensor of shape `(N, *) + O`
            action: Action tensor of shape `(N, *) + A`
            sample_shape: Shape to append to the samples of each model in the
                ensemble. Samples from all the models are then concatenated.

        Returns:
           Sample and log-likelihood tensors of shape `(N,) + S + (*,) + O` and
           `(N,) + S + (*,)` respectively, where `S` is the `sample_shape`.
        """
        outputs = [
            m.sample(obs[i], action[i], sample_shape) for i, m in enumerate(self)
        ]
        sample = torch.stack([s for s, _ in outputs])
        logp = torch.stack([p for _, p in outputs])
        return sample, logp

    @torch.jit.export
    def rsample(
        self, obs, action, sample_shape: List[int] = ()
    ) -> Tuple[Tensor, Tensor]:
        """Compute reparameterized samples and likelihoods for each model.

        Uses the same semantics as :meth:`StochasticModelEnsemble.sample`.
        """
        outputs = [
            m.rsample(obs[i], action[i], sample_shape) for i, m in enumerate(self)
        ]
        sample = torch.stack([s for s, _ in outputs])
        logp = torch.stack([p for _, p in outputs])
        return sample, logp

    @torch.jit.export
    def log_prob(self, obs, action, next_obs) -> Tensor:
        """Compute likelihoods for each model in the ensemble.

        Splits inputs by the first dimension in `N` chunks, where `N` is the
        ensemble size. Then, applies each model in the ensemble to one of the
        slices.

        Args:
            obs: Observation tensor of shape `(N, *) + O`
            action: Action tensor of shape `(N, *) + A`
            next_obs: Observation tensor of shape `(N, *) + O`

        Returns:
           Log-likelihood tensor of shape `(N, *)`
        """
        return torch.stack(
            [m.log_prob(obs[i], action[i], next_obs[i]) for i, m in enumerate(self)]
        )


class ForkedStochasticModelEnsemble(StochasticModelEnsemble):
    """Ensemble of stochastic models with parallelized methods."""

    # pylint:disable=abstract-method,protected-access

    @torch.jit.export
    def sample(self, obs, action, sample_shape: List[int] = ()):
        futures = [
            torch.jit._fork(m.sample, obs[i], action[i], sample_shape)
            for i, m in enumerate(self)
        ]
        outputs = [torch.jit._wait(f) for f in futures]
        sample = torch.stack([s for s, _ in outputs])
        logp = torch.stack([p for _, p in outputs])
        return sample, logp

    @torch.jit.export
    def rsample(self, obs, action, sample_shape: List[int] = ()):
        futures = [
            torch.jit._fork(m.rsample, obs[i], action[i], sample_shape)
            for i, m in enumerate(self)
        ]
        outputs = [torch.jit._wait(f) for f in futures]
        sample = torch.stack([s for s, _ in outputs])
        logp = torch.stack([p for _, p in outputs])
        return sample, logp

    @torch.jit.export
    def log_prob(self, obs, action, next_obs):
        futures = [
            torch.jit._fork(m.log_prob, obs[i], action[i], next_obs[i])
            for i, m in enumerate(self)
        ]
        return torch.stack([torch.jit._wait(f) for f in futures])



"""Constructors for stochastic dynamics models."""
from dataclasses import dataclass
from dataclasses import field

from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box

from .ensemble import ForkedStochasticModelEnsemble
from .ensemble import StochasticModelEnsemble
from .single import MLPModel
from .single import ResidualMLPModel

ModelSpec = MLPModel.spec_cls


@dataclass
class Spec(DataClassJsonMixin):
    """Specifications for stochastic dynamics model.

    Args:
        network: Specifications for stochastic model network
        input_dependent_scale: Whether to parameterize the Gaussian standard
            deviation as a function of the state and action
        residual: Whether to build model as a residual one, i.e., that
            predicts the change in state rather than the next state itself
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments. Used to initialize the model's Linear layers.
    """

    network: ModelSpec = field(default_factory=ModelSpec)
    input_dependent_scale: bool = True
    residual: bool = True
    initializer: dict = field(default_factory=dict)


def build(obs_space: Box, action_space: Box, spec: Spec) -> MLPModel:
    """Construct stochastic dynamics model.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for stochastic dynamics model

    Returns:
        A stochastic dynamics model
    """
    cls = ResidualMLPModel if spec.residual else MLPModel
    model = cls(obs_space, action_space, spec.network, spec.input_dependent_scale)
    model.initialize_parameters(spec.initializer)
    return model


@dataclass
class EnsembleSpec(Spec):
    """Specifications for stochastic dynamics model ensemble.

    Args:
        network: Specifications for stochastic model networks
        input_dependent_scale: Whether to parameterize the Gaussian standard
            deviation as a function of the state and action
        residual: Whether to build each model as a residual one, i.e., that
            predicts the change in state rather than the next state itself
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments. Used to initialize the models' Linear layers.
        ensemble_size: Number of models in the collection.
        parallelize: Whether to use an ensemble with parallelized `sample`,
            `rsample`, and `log_prob` methods
    """

    ensemble_size: int = 1
    parallelize: bool = False


def build_ensemble(
    obs_space: Box, action_space: Box, spec: EnsembleSpec
) -> StochasticModelEnsemble:
    """Construct stochastic dynamics model ensemble.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for stochastic dynamics model ensemble

    Returns:
        A stochastic dynamics model ensemble
    """
    models = [build(obs_space, action_space, spec) for _ in range(spec.ensemble_size)]
    cls = ForkedStochasticModelEnsemble if spec.parallelize else StochasticModelEnsemble
    ensemble = cls(models)
    return ensemble



"""SAC policy class using PyTorch."""
import torch
import torch.nn as nn
from ray.rllib.utils import override

from raylab.policy import TorchPolicy
from raylab.policy.action_dist import WrapStochasticPolicy
from raylab.policy.losses import MaximumEntropyDual
from raylab.policy.losses import ReparameterizedSoftPG
from raylab.policy.losses import SoftCDQLearning
from raylab.pytorch.nn.utils import update_polyak
from raylab.pytorch.optim import build_optimizer


class SACTorchPolicy(TorchPolicy):
    """Soft Actor-Critic policy in PyTorch to use with RLlib."""

    # pylint:disable=abstract-method
    dist_class = WrapStochasticPolicy

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)

        self._setup_actor_loss()
        self._setup_critic_loss()
        self._setup_alpha_loss()

    def _setup_actor_loss(self):
        self.loss_actor = ReparameterizedSoftPG(self.module.actor, self.module.critics)

    def _setup_critic_loss(self):
        self.loss_critic = SoftCDQLearning(
            self.module.critics, self.module.target_critics, self.module.actor
        )
        self.loss_critic.gamma = self.config["gamma"]

    def _setup_alpha_loss(self):
        action_space = self.action_space
        target_entropy = (
            -action_space.shape[0]
            if self.config["target_entropy"] == "auto"
            else self.config["target_entropy"]
        )
        self.loss_alpha = MaximumEntropyDual(
            self.module.alpha, self.module.actor.sample, target_entropy
        )

    @property
    @override(TorchPolicy)
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.sac import SACTrainer

        return SACTrainer.options

    @override(TorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["torch_optimizer"]

        components = "actor critics alpha".split()
        mapping = {
            name: build_optimizer(getattr(self.module, name), config[name])
            for name in components
        }

        optimizers.update(mapping)
        return optimizers

    @override(TorchPolicy)
    def learn_on_batch(self, samples):
        batch_tensors = self.lazy_tensor_dict(samples)
        info = {}

        alpha = self.module.alpha().item()
        self.loss_critic.alpha = alpha
        self.loss_actor.alpha = alpha

        info.update(self._update_critic(batch_tensors))
        info.update(self._update_actor(batch_tensors))
        if self.config["target_entropy"] is not None:
            info.update(self._update_alpha(batch_tensors))

        update_polyak(
            self.module.critics, self.module.target_critics, self.config["polyak"]
        )
        return info

    def _update_critic(self, batch_tensors):
        with self.optimizers.optimize("critics"):
            critic_loss, info = self.loss_critic(batch_tensors)
            critic_loss.backward()

        info.update(self.extra_grad_info("critics"))
        return info

    def _update_actor(self, batch_tensors):
        with self.optimizers.optimize("actor"):
            actor_loss, info = self.loss_actor(batch_tensors)
            actor_loss.backward()

        info.update(self.extra_grad_info("actor"))
        return info

    def _update_alpha(self, batch_tensors):
        with self.optimizers.optimize("alpha"):
            alpha_loss, info = self.loss_alpha(batch_tensors)
            alpha_loss.backward()

        info.update(self.extra_grad_info("alpha"))
        return info

    @torch.no_grad()
    def extra_grad_info(self, component):
        """Return statistics right after components are updated."""
        return {
            f"grad_norm({component})": nn.utils.clip_grad_norm_(
                getattr(self.module, component).parameters(), float("inf")
            ).item()
        }



"""Mixins to add environment mimicking functions."""
import inspect
import warnings
from typing import Optional
from typing import Union

import raylab.envs as envs
from raylab.utils.annotations import DynamicsFn
from raylab.utils.annotations import RewardFn
from raylab.utils.annotations import TerminationFn


class EnvFnMixin:
    """Add methods to create and set environment functions.

    Allows user to choose methods for setting the reward, termination,
    and dynamics functions, which may vary by experiment setup.

    Attributes:
        reward_fn: callable mapping dynamics triplets to reward tensors
        termination_fn: callable mapping dynamics triplets to termination
            tensors
        dynamics_fn: callable mapping (state, action) pairs to (next-state,
            log-probability) pairs
    """

    reward_fn: Optional[RewardFn]
    termination_fn: Optional[TerminationFn]
    dynamics_fn: Optional[DynamicsFn]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reward_fn = None
        self.termination_fn = None
        self.dynamics_fn = None

    def set_reward_from_config(self):
        """Build and set the reward function from environment configurations."""
        env_id, env_config = self.config["env"], self.config["env_config"]
        self.reward_fn = envs.get_reward_fn(env_id, env_config)
        self._set_reward_hook()

    def set_reward_from_callable(self, function: RewardFn):
        """Set the reward function from an external callable.

        Args:
            function: A callable that reproduces the environment's reward
                function
        """
        self._check_not_instance_method(function)
        self.reward_fn = function
        self._set_reward_hook()

    def set_termination_from_config(self):
        """Build and set a termination function from environment configurations."""
        env_id, env_config = self.config["env"], self.config["env_config"]
        self.termination_fn = envs.get_termination_fn(env_id, env_config)
        self._set_termination_hook()

    def set_termination_from_callable(self, function: TerminationFn):
        """Set the termination function from an external callable.

        Args:
            function: A callable that reproduces the environment's termination
                function
        """
        self._check_not_instance_method(function)
        self.termination_fn = function
        self._set_termination_hook()

    def set_dynamics_from_callable(self, function: DynamicsFn):
        """Set the dynamics function from an external callable.

        Args:
            function: A callable that reproduces the environment's dynamics
                function
        """
        self._check_not_instance_method(function)
        self.dynamics_fn = function
        self._set_dynamics_hook()

    def _set_reward_hook(self):
        """Procedure to run upon setting the reward function.

        Subclasses should override this to do extra processing whenever
        `set_reward_from_callable` or `set_reward_from_config` is called.
        For example, the subclass may set up losses which assume access to
        environment functions.
        """

    def _set_termination_hook(self):
        """Procedure to run upon setting the termination function.

        Subclasses should override this to do extra processing whenever
        `set_termination_from_callable` or `set_termination_from_config` is
        called. For example, the subclass may set up losses which assume access
        to environment functions.
        """

    def _set_dynamics_hook(self):
        """Procedure to run upon setting the dynamics function.

        Subclasses should override this to do extra processing whenever
        `set_dynamics_from_callable` is called. For example, the subclass may
        set up losses which assume access to environment functions.
        """

    @staticmethod
    def _check_not_instance_method(env_fn: Union[RewardFn, TerminationFn]):
        if inspect.ismethod(env_fn):
            warnings.warn(
                f"{env_fn.__name__} function is a bound instance method of"
                f" {env_fn.__self__}. Prefer static environment functions"
                " with markovian states that are not bounded to a particular"
                " instance"
            )



"""Model-based architecture with Normalizing Flows."""

from .abstract import AbstractModelActorCritic
from .mixins import MaximumEntropyMixin
from .mixins import NormalizingFlowActorMixin
from .mixins import NormalizingFlowModelMixin
from .mixins import StateValueMixin


# pylint:disable=abstract-method,too-many-ancestors
class NFMBRL(
    MaximumEntropyMixin,
    NormalizingFlowModelMixin,
    NormalizingFlowActorMixin,
    StateValueMixin,
    AbstractModelActorCritic,
):
    """
    Module architecture with normalizing flow actor and model, and state value function.
    """



"""Continuous Q-Learning with Normalized Advantage Functions."""
from raylab.agents import trainer
from raylab.agents.off_policy import OffPolicyTrainer

from .policy import NAFTorchPolicy


@trainer.configure
@trainer.option("torch_optimizer/type", "Adam")
@trainer.option("torch_optimizer/lr", 3e-4)
@trainer.option(
    "polyak",
    0.995,
    help="Interpolation factor in polyak averaging for target networks.",
)
@trainer.option("module/type", "NAF")
@trainer.option("module/separate_behavior", True)
@trainer.option(
    "exploration_config/type", "raylab.utils.exploration.ParameterNoise", override=True
)
@trainer.option(
    "exploration_config/param_noise_spec",
    {"initial_stddev": 0.1, "desired_action_stddev": 0.2, "adaptation_coeff": 1.01},
)
@trainer.option("exploration_config/pure_exploration_steps", 1000)
@trainer.option("evaluation_config/explore", False, override=True)
class NAFTrainer(OffPolicyTrainer):
    """Single agent trainer for NAF."""

    _name = "NAF"
    _policy = NAFTorchPolicy



"""NAF policy class using PyTorch."""
import torch
import torch.nn as nn
from ray.rllib.utils import override

from raylab.policy import TorchPolicy
from raylab.policy.action_dist import WrapDeterministicPolicy
from raylab.policy.losses import ClippedDoubleQLearning
from raylab.pytorch.nn.utils import update_polyak
from raylab.pytorch.optim import build_optimizer


class NAFTorchPolicy(TorchPolicy):
    """Normalized Advantage Function policy in Pytorch to use with RLlib."""

    # pylint:disable=abstract-method
    dist_class = WrapDeterministicPolicy

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loss_fn = ClippedDoubleQLearning(
            self.module.critics, self.module.target_vcritics
        )
        self.loss_fn.gamma = self.config["gamma"]

    @property
    @override(TorchPolicy)
    def options(self):
        # pylint:disable=cyclic-import
        from raylab.agents.naf import NAFTrainer

        return NAFTrainer.options

    @override(TorchPolicy)
    def _make_module(self, obs_space, action_space, config):
        module_config = config["module"]
        module_config["type"] = "NAF"
        # pylint:disable=no-member
        return super()._make_module(obs_space, action_space, config)

    @override(TorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        optimizers.update(
            naf=build_optimizer(self.module.critics, self.config["torch_optimizer"])
        )
        return optimizers

    @override(TorchPolicy)
    def learn_on_batch(self, samples):
        batch_tensors = self.lazy_tensor_dict(samples)
        with self.optimizers.optimize("naf"):
            loss, info = self.loss_fn(batch_tensors)
            loss.backward()

        info.update(self.extra_grad_info())
        update_polyak(
            self.module.vcritics, self.module.target_vcritics, self.config["polyak"]
        )
        return info

    @torch.no_grad()
    def extra_grad_info(self):
        """Compute gradient norm for components."""
        return {
            "grad_norm": nn.utils.clip_grad_norm_(
                self.module.critics.parameters(), float("inf")
            ).item()
        }



"""Trainer and configuration for SVG(1)."""
from ray.rllib import SampleBatch
from ray.rllib.utils import override

from raylab.agents import trainer
from raylab.agents.model_based import set_policy_with_env_fn
from raylab.agents.off_policy import OffPolicyTrainer
from raylab.utils.replay_buffer import ReplayField

from .policy import SVGOneTorchPolicy


@trainer.configure
@trainer.option(
    "torch_optimizer/type", "Adam", help="Optimizer type for model, actor, and critic"
)
@trainer.option("torch_optimizer/model", {"lr": 1e-3})
@trainer.option("torch_optimizer/actor", {"lr": 1e-3})
@trainer.option("torch_optimizer/critic", {"lr": 1e-3})
@trainer.option(
    "vf_loss_coeff",
    1.0,
    help="Weight of the fitted V loss in the joint model-value loss",
)
@trainer.option("max_grad_norm", 10.0, help="Clip gradient norms by this value")
@trainer.option(
    "max_is_ratio", 5.0, help="Clip importance sampling weights by this value"
)
@trainer.option(
    "polyak",
    0.995,
    help="Interpolation factor in polyak averaging for target networks.",
)
@trainer.option(
    "kl_schedule",
    {"initial_coeff": 0},
    help="Options for adaptive KL coefficient. See raylab.utils.adaptive_kl",
)
@trainer.option(
    "replay_kl",
    True,
    help="""
    Whether to penalize KL divergence with the current policy or past policies
    that generated the replay pool.
    """,
)
@trainer.option("module", {"type": "SVGModule-v0"}, override=True)
@trainer.option(
    "exploration_config/type", "raylab.utils.exploration.StochasticActor", override=True
)
@trainer.option("exploration_config/pure_exploration_steps", 1000)
@trainer.option("evaluation_config/explore", False)
class SVGOneTrainer(OffPolicyTrainer):
    """Single agent trainer for SVG(1)."""

    # pylint:disable=attribute-defined-outside-init
    _name = "SVG(1)"
    _policy = SVGOneTorchPolicy

    @override(OffPolicyTrainer)
    def _init(self, config, env_creator):
        super()._init(config, env_creator)
        set_policy_with_env_fn(self.workers, fn_type="reward")

    @override(OffPolicyTrainer)
    def build_replay_buffer(self, config):
        super().build_replay_buffer(config)
        self.replay.add_fields(ReplayField(SampleBatch.ACTION_LOGP))

    @override(OffPolicyTrainer)
    def _before_replay_steps(self, policy):
        if not self.config["replay_kl"]:
            policy.update_old_policy()



"""PyTorch related utilities."""
from typing import Dict
from typing import Iterator
from typing import Union

import numpy as np
import torch
from torch.autograd import grad


def flat_grad(
    outputs: Union[torch.Tensor, Iterator[torch.Tensor]],
    inputs: Iterator[torch.Tensor],
    *args,
    **kwargs
) -> torch.Tensor:
    """Compute gradients and return a flattened array."""
    params = list(inputs)
    grads = grad(outputs, params, *args, **kwargs)
    zeros = torch.zeros
    return torch.cat(
        [zeros(p.numel()) if g is None else g.flatten() for p, g in zip(params, grads)]
    )


def convert_to_tensor(arr, device: torch.device) -> torch.Tensor:
    """Convert array-like object to tensor and cast it to appropriate device.

    Arguments:
        arr (array_like): object which can be converted using `np.asarray`
        device: device to cast the resulting tensor to

    Returns:
        The array converted to a `torch.Tensor`.
    """
    if torch.is_tensor(arr):
        return arr.to(device)
    tensor = torch.from_numpy(np.asarray(arr))
    if tensor.dtype == torch.double:
        tensor = tensor.float()
    return tensor.to(device)


class TensorDictDataset(torch.utils.data.Dataset):
    """Dataset wrapping a dict of tensors.

    Args:
        tensor_dict: dictionary mapping strings to tensors
    """

    def __init__(self, tensor_dict: Dict[str, torch.Tensor]):
        super().__init__()
        batch_size = next(iter(tensor_dict.values())).size(0)
        assert all(tensor.size(0) == batch_size for tensor in tensor_dict.values())
        self.tensor_dict = tensor_dict

    def __getitem__(self, index: int) -> Dict[str, torch.Tensor]:
        return {k: v[index] for k, v in self.tensor_dict.items()}

    def __len__(self) -> int:
        return next(iter(self.tensor_dict.values())).size(0)



# pylint:disable=missing-docstring
import torch.nn as nn
from ray.rllib.utils import override


class Lambda(nn.Module):
    """Neural network module that stores and applies a function on inputs."""

    def __init__(self, func):
        super().__init__()
        self.func = func

    @override(nn.Module)
    def forward(self, inputs):  # pylint:disable=arguments-differ
        return self.func(inputs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['raylab',
 'raylab.agents',
 'raylab.agents.acktr',
 'raylab.agents.mage',
 'raylab.agents.mapo',
 'raylab.agents.mapo.dapo',
 'raylab.agents.mapo.mle',
 'raylab.agents.mbpo',
 'raylab.agents.naf',
 'raylab.agents.sac',
 'raylab.agents.sac.dyna',
 'raylab.agents.sop',
 'raylab.agents.svg',
 'raylab.agents.svg.inf',
 'raylab.agents.svg.one',
 'raylab.agents.svg.soft',
 'raylab.agents.trpo',
 'raylab.cli',
 'raylab.envs',
 'raylab.envs.environments',
 'raylab.envs.wrappers',
 'raylab.logger',
 'raylab.policy',
 'raylab.policy.losses',
 'raylab.policy.model_based',
 'raylab.policy.modules',
 'raylab.policy.modules.actor',
 'raylab.policy.modules.actor.policy',
 'raylab.policy.modules.critic',
 'raylab.policy.modules.model',
 'raylab.policy.modules.model.stochastic',
 'raylab.policy.modules.networks',
 'raylab.policy.modules.v0',
 'raylab.policy.modules.v0.mixins',
 'raylab.pytorch',
 'raylab.pytorch.distributions',
 'raylab.pytorch.nn',
 'raylab.pytorch.nn.distributions',
 'raylab.pytorch.nn.distributions.flows',
 'raylab.pytorch.nn.modules',
 'raylab.pytorch.optim',
 'raylab.tune',
 'raylab.utils',
 'raylab.utils.exploration']

package_data = \
{'': ['*']}

install_requires = \
['bokeh>=2.1.0,<3.0.0',
 'click>=7.1.2,<8.0.0',
 'dataclasses-json>=0.5.1,<0.6.0',
 'opencv-python>=4.2.0,<5.0.0',
 'poetry-version>=0.1.5,<0.2.0',
 'ray[rllib,dashboard]>=0.8.5,<0.9.0',
 'streamlit>=0.62,<0.65',
 'torch>=1.5.1,<2.0.0']

extras_require = \
{':python_version >= "3.7" and python_version < "4.0"': ['cachetools>=4.1.0,<5.0.0']}

entry_points = \
{'console_scripts': ['raylab = raylab.cli:raylab']}

setup_kwargs = {
    'name': 'raylab',
    'version': '0.11.21',
    'description': 'Reinforcement learning algorithms in RLlib and PyTorch.',
    'long_description': '======\nraylab\n======\n\n|PyPI| |Tests| |Dependabot| |License| |CodeStyle|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/raylab?logo=PyPi&logoColor=white&color=blue\n\t  :alt: PyPI\n\n.. |Tests| image:: https://img.shields.io/github/workflow/status/angelolovatto/raylab/Poetry%20package?label=tests&logo=GitHub\n\t   :alt: GitHub Workflow Status\n\n.. |Dependabot| image:: https://api.dependabot.com/badges/status?host=github&repo=angelolovatto/raylab\n\t\t:target: https://dependabot.com\n\n.. |License| image:: https://img.shields.io/github/license/angelolovatto/raylab?color=blueviolet&logo=github\n\t     :alt: GitHub\n\n.. |CodeStyle| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n\t       :target: https://github.com/psf/black\n\n\nReinforcement learning algorithms in `RLlib <https://github.com/ray-project/ray/tree/master/rllib>`_ and `PyTorch <https://pytorch.org>`_.\n\n\nIntroduction\n------------\n\nRaylab provides agents and environments to be used with a normal RLlib/Tune setup.\n\n.. code-block:: python\n\n             import ray\n             from ray import tune\n             import raylab\n\n             def main():\n                 raylab.register_all_agents()\n                 raylab.register_all_environments()\n                 ray.init()\n                 tune.run(\n                     "NAF",\n                     local_dir=...,\n                     stop={"timesteps_total": 100000},\n                     config={\n                         "env": "CartPoleSwingUp-v0",\n                         "exploration_config": {\n                             "type": tune.grid_search([\n                                 "raylab.utils.exploration.GaussianNoise",\n                                 "raylab.utils.exploration.ParameterNoise"\n                             ])\n                         }\n                         ...\n                     },\n                 )\n\n             if __name__ == "__main__":\n                 main()\n\n\nOne can then visualize the results using `raylab dashboard`\n\n.. image:: https://i.imgur.com/bVc6WC5.png\n        :align: center\n\n\nInstallation\n------------\n\n.. code:: bash\n\n          pip install raylab\n\n\nAlgorithms\n----------\n\n+--------------------------------------------------------+-------------------------+\n| Paper                                                  | Agent Name              |\n+--------------------------------------------------------+-------------------------+\n| `Actor Critic using Kronecker-factored Trust Region`_  | ACKTR                   |\n+--------------------------------------------------------+-------------------------+\n| `Trust Region Policy Optimization`_                    | TRPO                    |\n+--------------------------------------------------------+-------------------------+\n| `Normalized Advantage Function`_                       | NAF                     |\n+--------------------------------------------------------+-------------------------+\n| `Stochastic Value Gradients`_                          | SVG(inf)/SVG(1)/SoftSVG |\n+--------------------------------------------------------+-------------------------+\n| `Soft Actor-Critic`_                                   | SoftAC                  |\n+--------------------------------------------------------+-------------------------+\n| `Streamlined Off-Policy`_ (DDPG)                       | SOP                     |\n+--------------------------------------------------------+-------------------------+\n| `Model-Based Policy Optimization`_                     | MBPO                    |\n+--------------------------------------------------------+-------------------------+\n| `Model-based Action-Gradient-Estimator`_               | MAGE                    |\n+--------------------------------------------------------+-------------------------+\n\n\n.. _`Actor Critic using Kronecker-factored Trust Region`: https://arxiv.org/abs/1708.05144\n.. _`Trust Region Policy Optimization`: http://proceedings.mlr.press/v37/schulman15.html\n.. _`Normalized Advantage Function`: http://proceedings.mlr.press/v48/gu16.html\n.. _`Stochastic Value Gradients`: http://papers.nips.cc/paper/5796-learning-continuous-control-policies-by-stochastic-value-gradients\n.. _`Soft Actor-Critic`: http://proceedings.mlr.press/v80/haarnoja18b.html\n.. _`Model-Based Policy Optimization`: http://arxiv.org/abs/1906.08253\n.. _`Streamlined Off-Policy`: https://arxiv.org/abs/1910.02208\n.. _`Model-based Action-Gradient-Estimator`: https://arxiv.org/abs/2004.14309\n\n\nCommand-line interface\n----------------------\n\n.. role:: bash(code)\n   :language: bash\n\nFor a high-level description of the available utilities, run :bash:`raylab --help`\n\n.. code:: bash\n\n\tUsage: raylab [OPTIONS] COMMAND [ARGS]...\n\n\t  RayLab: Reinforcement learning algorithms in RLlib.\n\n\tOptions:\n\t  --help  Show this message and exit.\n\n\tCommands:\n\t  dashboard    Launch the experiment dashboard to monitor training progress.\n\t  episodes     Launch the episode dashboard to monitor state and action...\n\t  experiment   Launch a Tune experiment from a config file.\n\t  find-best    Find the best experiment checkpoint as measured by a metric.\n\t  info         View information about an agent\'s config parameters.\n\t  rollout      Wrap `rllib rollout` with customized options.\n\t  test-module  Launch dashboard to test generative models from a checkpoint.\n\n\nPackages\n--------\n\nThe project is structured as follows\n::\n\n    raylab\n    |-- agents            # Trainer and Policy classes\n    |-- cli               # Command line utilities\n    |-- envs              # Gym environment registry and utilities\n    |-- logger            # Tune loggers\n    |-- policy            # Extensions and customizations of RLlib\'s policy API\n    |   |-- losses        # RL loss functions\n    |   |-- modules       # PyTorch neural network modules for TorchPolicy\n    |-- pytorch           # PyTorch extensions\n    |-- utils             # miscellaneous utilities\n',
    'author': 'Ângelo Gregório Lovatto',
    'author_email': 'angelolovatto@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/angelolovatto/raylab',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""SVG Architecture with disjoint model, actor, and critic."""
from raylab.utils.dictionaries import deep_merge

from .abstract import AbstractModelActorCritic
from .mixins import StateValueMixin
from .mixins import StochasticActorMixin
from .mixins import SVGModelMixin


BASE_CONFIG = {
    "replay_kl": False,
    "actor": {
        "encoder": {"units": (100, 100), "activation": "Tanh"},
        "initializer_options": {"name": "xavier_uniform"},
        "input_dependent_scale": False,
    },
    "critic": {
        "target_vf": True,
        "encoder": {"units": (400, 200), "activation": "Tanh"},
        "initializer_options": {"name": "xavier_uniform"},
    },
    "model": {
        "residual": True,
        "input_dependent_scale": False,
        "encoder": {"units": (40, 40), "activation": "Tanh", "delay_action": True},
        "initializer_options": {"name": "xavier_uniform"},
    },
}


class SVGModule(
    SVGModelMixin, StochasticActorMixin, StateValueMixin, AbstractModelActorCritic,
):
    """Module architecture with reparameterized actor and model.

    Allows inference of noise variables given existing samples.
    """

    # pylint:disable=abstract-method

    def __init__(self, obs_space, action_space, config):
        config = deep_merge(BASE_CONFIG, config, False, ["actor", "critic", "model"])
        super().__init__(obs_space, action_space, config)

        if config.get("replay_kl") is False:
            old = self._make_actor(obs_space, action_space, config)
            self.old_actor = old["actor"].requires_grad_(False)



"""Parameterized action-value estimators."""
from abc import ABC
from abc import abstractmethod

import torch
import torch.nn as nn
from gym.spaces import Box
from torch import Tensor

from raylab.policy.modules.networks.mlp import StateActionMLP


MLPSpec = StateActionMLP.spec_cls


class QValue(ABC, nn.Module):
    """Neural network module emulating a Q value function."""

    # pylint:disable=arguments-differ
    @abstractmethod
    def forward(self, obs: Tensor, action: Tensor) -> Tensor:
        """Main forward pass mapping obs and actions to Q-values.

        Note:
            The output tensor has a last singleton dimension, i.e., for a batch
            of 10 obs-action pairs, the output will have shape (10, 1).
        """


class MLPQValue(QValue):
    """Q-value function with a multilayer perceptron encoder.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Multilayer perceptron specifications

    Attributes:
        encoder: NN module mapping states to 1D features. Must have an
            `out_features` attribute with the size of the output features
    """

    spec_cls = MLPSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: MLPSpec):
        super().__init__()
        self.encoder = StateActionMLP(obs_space, action_space, spec)
        self.value_linear = nn.Linear(self.encoder.out_features, 1)

    def forward(self, obs: Tensor, action: Tensor) -> Tensor:
        features = self.encoder(obs, action)
        return self.value_linear(features)

    def initialize_parameters(self, initializer_spec: dict):
        """Initialize all Linear models in the encoder.

        Uses `raylab.pytorch.nn.init.initialize_` to create an initializer
        function.

        Args:
            initializer_spec: Dictionary with mandatory `name` key corresponding
                to the initializer function name in `torch.nn.init` and optional
                keyword arguments.
        """
        self.encoder.initialize_parameters(initializer_spec)


class QValueEnsemble(nn.ModuleList, QValue):
    """A static list of Q-value estimators.

    Args:
        q_values: A list of QValue modules
    """

    def __init__(self, q_values):
        cls_name = type(self).__name__
        assert all(
            isinstance(q, QValue) for q in q_values
        ), f"All modules in {cls_name} must be instances of QValue."
        super().__init__(q_values)

    def forward(self, obs: Tensor, action: Tensor) -> Tensor:
        """Evaluate each Q estimator in the ensemble.

        Args:
            obs: The observation tensor
            action: The action tensor

        Returns:
            A tensor of shape `(*, N)`, where `N` is the ensemble size
        """
        # pylint:disable=arguments-differ
        return self._action_values(obs, action)

    def _action_values(self, obs: Tensor, act: Tensor) -> Tensor:
        return torch.cat([m(obs, act) for m in self], dim=-1)

    def initialize_parameters(self, initializer_spec: dict):
        """Initialize each Q estimator in the ensemble.

        Args:
            initializer_spec: Dictionary with mandatory `name` key corresponding
                to the initializer function name in `torch.nn.init` and optional
                keyword arguments.
        """
        for q_value in self:
            q_value.initialize_parameters(initializer_spec)


class ForkedQValueEnsemble(QValueEnsemble):
    """Ensemble of Q-value estimators with parallelized forward pass."""

    def _action_values(self, obs: Tensor, act: Tensor) -> Tensor:
        # pylint:disable=protected-access
        futures = [torch.jit._fork(m, obs, act) for m in self]
        return torch.cat([torch.jit._wait(f) for f in futures], dim=-1)



"""Parameterized normalized advantage function estimators."""
import torch.nn as nn
from gym.spaces import Box
from torch import Tensor

import raylab.pytorch.nn as nnx
from raylab.policy.modules.actor.policy.deterministic import DeterministicPolicy

from .q_value import QValue
from .v_value import VValue


class NAFQValue(QValue):
    """Neural network module emulating a Normalized Advantage Function.

    Args:
        action_space: Continuous action space
        policy: Continuous deterministic policy

    Attributes:
        policy: Deterministic policy
        v_value: State-value function
    """

    def __init__(self, action_space: Box, policy: DeterministicPolicy):
        super().__init__()
        act_size = action_space.shape[0]
        self.policy = policy

        self._value_linear = nn.Linear(policy.encoder.out_features, 1)
        self.v_value = NAFVValue(policy.encoder, self._value_linear)

        self._tril = nnx.TrilMatrix(policy.encoder.out_features, act_size)
        self._advantage = AdvantageFunction(
            self._tril, nn.Sequential(policy.action_linear, policy.squashing)
        )

    def forward(self, obs: Tensor, action: Tensor) -> Tensor:
        logits = self.policy.encoder(obs)
        best_value = self._value_linear(logits)
        advantage = self._advantage(logits, action)
        return advantage + best_value


class NAFVValue(VValue):
    """Wrapper around NAF's state-value function."""

    def __init__(self, encoder: nn.Module, value_linear: nn.Linear):
        super().__init__()
        self.encoder = encoder
        self.value_linear = value_linear

    def forward(self, obs: Tensor) -> Tensor:
        return self.value_linear(self.encoder(obs))


class AdvantageFunction(nn.Module):
    """Neural network module implementing the advantage function term of NAF."""

    def __init__(self, tril_module, action_module):
        super().__init__()
        self.tril_module = tril_module
        self.action_module = action_module

    def forward(self, logits, action):  # pylint:disable=arguments-differ
        tril_matrix = self.tril_module(logits)  # square matrix [..., N, N]
        best_action = self.action_module(logits)  # batch of actions [..., N]
        action_diff = (action - best_action).unsqueeze(-1)  # column vector [..., N, 1]
        vec = tril_matrix.matmul(action_diff)  # column vector [..., N, 1]
        advantage = -0.5 * vec.transpose(-1, -2).matmul(vec).squeeze(-1)
        return advantage



"""Mixins for loss functions."""
from dataclasses import dataclass
from typing import Optional
from typing import Tuple

import numpy as np
import torch
from torch import Tensor

from raylab.utils.annotations import RewardFn
from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict
from raylab.utils.annotations import TerminationFn


@dataclass
class EnvFunctions:
    """Collection of environment emulating functions."""

    reward: Optional[RewardFn] = None
    termination: Optional[TerminationFn] = None

    @property
    def initialized(self):
        """Whether or not all functions are set."""
        return self.reward is not None and self.termination is not None


class EnvFunctionsMixin:
    """Adds private, externally set environment functions.

    The resulting loss function will have an `_env` private attribute. The user
    can use the `set_reward_fn` and `set_termination_fn` methods to set the
    environment functions. When both are set, `_env.initialized` will be True.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._env = EnvFunctions()

    def set_reward_fn(self, function: RewardFn):
        """Set reward function to provided callable."""
        self._env.reward = function

    def set_termination_fn(self, function: TerminationFn):
        """Set termination function to provided callable."""
        self._env.termination = function


class UniformModelPriorMixin:
    """Add methods for using model ensembles with uniform prior distribution.

    Expects a model ensemble as a `_models` instance attribute.

    Attributes:
        grad_estimator: Gradient estimator for expecations ('PD' or 'SF')
        model_samples: Number of next states to draw from the model
    """

    grad_estimator: str = "SF"
    model_samples: int = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._rng = np.random.default_rng()

    def seed(self, seed: int):
        """Seeds the RNG for choosing a model from the ensemble."""
        self._rng = np.random.default_rng(seed)

    def transition(self, obs: Tensor, action: Tensor) -> Tuple[Tensor, Tensor]:
        """Compute virtual transition and its log density.

        Samples a model from the ensemble using the internal RNG and uses it to
        generate the next state.

        Args:
            obs: The current state
            action: The action sampled from the stochastic policy

        Returns:
            A tuple with the next state and its log-likelihood generated from
            a model sampled from the ensemble
        """
        sample_shape = (self.model_samples,)

        model = self._rng.choice(self._models)
        dist_params = model.params(obs, action)
        if self.grad_estimator == "SF":
            next_obs, logp = model.dist.sample(dist_params, sample_shape)
        elif self.grad_estimator == "PD":
            next_obs, logp = model.dist.rsample(dist_params, sample_shape)
        return next_obs, logp, dist_params

    @staticmethod
    def model_dist_info(dist_params: TensorDict) -> StatDict:
        """Computes average, min, and max of model distribution parameters.

        Args:
            dist_params: Dictionary mapping names to dynamics model distribution
                parameters

        Returns:
            Dictionary with average, minimum, and maximum of each parameter as
                floats
        """
        info = {}
        info.update({"min_model_" + k: v.min().item() for k, v in dist_params.items()})
        info.update(
            {"mean_model_" + k: v.mean().item() for k, v in dist_params.items()}
        )
        info.update({"max_model_" + k: v.max().item() for k, v in dist_params.items()})
        return info

    def verify_model(self, obs: Tensor, act: Tensor):
        """Verify model suitability for the current gradient estimator.

        Assumes all models in the ensemble behave the same way.

        Args:
            obs: Dummy observation tensor
            act: Dummy action tensor

        Raises:
            AssertionError: If the internal model does not satisfy requirements
                for gradient estimation
        """
        model = self._models[0]
        if self.grad_estimator == "SF":
            sample, logp = model.sample(obs, act.requires_grad_())
            assert sample.grad_fn is None
            assert logp is not None
            logp.mean().backward()
            assert (
                act.grad is not None
            ), "Transition grad log_prob must exist for SF estimator"
            assert not torch.allclose(act.grad, torch.zeros_like(act))
        if self.grad_estimator == "PD":
            sample, _ = model.rsample(obs.requires_grad_(), act.requires_grad_())
            sample.mean().backward()
            assert (
                act.grad is not None
            ), "Transition grad w.r.t. state and action must exist for PD estimator"
            assert not torch.allclose(act.grad, torch.zeros_like(act))



# pylint:disable=all
"""
Reference:

Glow: Generative Flow with Invertible 1x1 Convolutions, Kingma and Dhariwal, Jul 2018
https://arxiv.org/abs/1807.03039
"""
import torch
from torch import nn


class Invertible1x1Conv(nn.Module):
    """
    As introduced in Glow paper.
    """

    def __init__(self, dim):
        super().__init__()
        self.dim = dim
        Q = torch.nn.init.orthogonal_(torch.randn(dim, dim))
        P, L, U = torch.lu_unpack(*Q.lu())
        self.P = P  # remains fixed during optimization
        self.L = nn.Parameter(L)  # lower triangular portion
        self.S = nn.Parameter(U.diag())  # "crop out" the diagonal to its own parameter
        self.U = nn.Parameter(
            torch.triu(U, diagonal=1)
        )  # "crop out" diagonal, stored in S

    def _assemble_W(self):
        """ assemble W from its pieces (P, L, U, S) """
        L = torch.tril(self.L, diagonal=-1) + torch.diag(torch.ones(self.dim))
        U = torch.triu(self.U, diagonal=1)
        W = self.P @ L @ (U + torch.diag(self.S))
        return W

    def forward(self, x):
        W = self._assemble_W()
        z = x @ W
        log_abs_det_jacobian = torch.sum(torch.log(torch.abs(self.S)))
        return z, log_abs_det_jacobian

    def backward(self, z):
        W = self._assemble_W()
        W_inv = torch.inverse(W)
        x = z @ W_inv
        log_abs_det_jacobian = -torch.sum(torch.log(torch.abs(self.S)))
        return x, log_abs_det_jacobian



"""Modularized Q-Learning procedures."""
from abc import ABC
from abc import abstractmethod
from typing import Optional
from typing import Tuple
from typing import Union

import torch
import torch.nn as nn
from ray.rllib import SampleBatch
from torch import Tensor

import raylab.utils.dictionaries as dutil
from raylab.policy.modules.actor.policy.deterministic import DeterministicPolicy
from raylab.policy.modules.actor.policy.stochastic import StochasticPolicy
from raylab.policy.modules.critic.q_value import QValue
from raylab.policy.modules.critic.v_value import PolicyQValue
from raylab.policy.modules.critic.v_value import VValue
from raylab.policy.modules.model.stochastic.ensemble import StochasticModelEnsemble
from raylab.utils.annotations import StatDict
from raylab.utils.annotations import TensorDict

from .abstract import Loss
from .mixins import EnvFunctionsMixin
from .mixins import UniformModelPriorMixin


class QLearningMixin(ABC):
    """Adds default call for Q-Learning losses."""

    # pylint:disable=too-few-public-methods
    batch_keys = (
        SampleBatch.CUR_OBS,
        SampleBatch.ACTIONS,
        SampleBatch.REWARDS,
        SampleBatch.NEXT_OBS,
        SampleBatch.DONES,
    )

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, TensorDict]:
        """Compute loss for Q-value function."""
        obs, actions, rewards, next_obs, dones = dutil.get_keys(batch, *self.batch_keys)
        with torch.no_grad():
            target_values = self.critic_targets(rewards, next_obs, dones)
        loss_fn = nn.MSELoss()
        values = self.critics(obs, actions)
        critic_loss = loss_fn(values, target_values)

        stats = {"loss(critics)": critic_loss.item()}
        stats.update(self.q_value_info(values))
        return critic_loss, stats

    @abstractmethod
    def critic_targets(
        self, rewards: Tensor, next_obs: Tensor, dones: Tensor
    ) -> Tensor:
        """Compute clipped 1-step approximation of Q^{\\pi}(s, a)."""

    @staticmethod
    def q_value_info(value: Tensor) -> StatDict:
        """Return the average, min, and max Q-values in a batch."""
        info = {
            "q_mean": value.mean().item(),
            "q_max": value.max().item(),
            "q_min": value.min().item(),
        }
        return info


class ClippedDoubleQLearning(QLearningMixin, Loss):
    """Clipped Double Q-Learning.

    Use the minimun of two target value functions as the next state-value in
    the target for fitted Q iteration.

    Args:
        critics: Main action-values
        target_critics: Target value functions. If provided an action-value
            function, requires a deterministic policy as the `actor` argument.
        actor: Optional deterministic policy for the next state

    Attributes:
        gamma: discount factor
    """

    gamma: float = 0.99

    def __init__(
        self,
        critics: QValue,
        target_critics: Union[QValue, VValue],
        actor: Optional[DeterministicPolicy] = None,
    ):
        self.critics = critics
        if isinstance(target_critics, QValue):
            if actor is None:
                raise ValueError(
                    f"Passing a Q-value function to {type(self).__name__}"
                    " requires a deterministic policy as the `actor` argument."
                )
            self.target_critics = PolicyQValue(policy=actor, q_value=target_critics)
        else:
            self.target_critics = target_critics

    def critic_targets(self, rewards, next_obs, dones):
        unclipped_values = self.target_critics(next_obs)
        values, _ = unclipped_values.min(dim=-1)
        next_vals = torch.where(dones, torch.zeros_like(values), values)
        target = rewards + self.gamma * next_vals
        return target.unsqueeze(-1).expand_as(unclipped_values)


class SoftCDQLearning(QLearningMixin, Loss):
    """Clipped Double Q-Learning for maximum entropy RL.

    Args:
        critics: Main action-values
        target_critics: Target action-values
        actor: Stochastic policy for the next state

    Attributes:
        gamma: discount factor
        alpha: entropy coefficient
    """

    gamma: float = 0.99
    alpha: float = 0.05

    def __init__(
        self, critics: QValue, target_critics: QValue, actor: StochasticPolicy,
    ):
        self.critics = critics
        self.target_critics = target_critics
        self.actor = actor

    def critic_targets(self, rewards, next_obs, dones):
        next_acts, next_logp = self.actor.sample(next_obs)
        unclipped_values = self.target_critics(next_obs, next_acts)
        values, _ = unclipped_values.min(dim=-1)

        next_vals = torch.where(dones, torch.zeros_like(values), values)
        next_entropy = torch.where(dones, torch.zeros_like(next_logp), -next_logp)
        target = rewards + self.gamma * (next_vals + self.alpha * next_entropy)
        return target.unsqueeze(-1).expand_as(unclipped_values)


class DynaSoftCDQLearning(EnvFunctionsMixin, UniformModelPriorMixin, SoftCDQLearning):
    """Loss function Dyna-augmented soft clipped double Q-learning.

    Args:
        critics: Main action-values
        models: Stochastic model ensemble
        target_critics: Target action-values
        actor: Stochastic policy for the next state

    Attributes:
        gamma: discount factor
        alpha: entropy coefficient
    """

    batch_keys: Tuple[str] = (SampleBatch.CUR_OBS,)

    def __init__(
        self,
        critics: QValue,
        models: StochasticModelEnsemble,
        target_critics: QValue,
        actor: StochasticPolicy,
    ):
        super().__init__(critics, target_critics, actor)
        self._models = models

    @property
    def initialized(self) -> bool:
        """Whether or not the loss function has all the necessary components."""
        return self._env.initialized

    def transition(self, obs, action):
        next_obs, logp, dist_params = super().transition(obs, action)
        # Squeeze the model samples dimension
        return next_obs.squeeze(dim=0), logp.squeeze(dim=0), dist_params

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        assert self.initialized, (
            "Environment functions missing. "
            "Did you set reward and termination functions?"
        )
        obs = batch[SampleBatch.CUR_OBS]
        action, _ = self.actor.sample(obs)
        next_obs, _, dist_params = self.transition(obs, action)

        reward = self._env.reward(obs, action, next_obs)
        done = self._env.termination(obs, action, next_obs)

        loss_fn = nn.MSELoss()
        value = self.critics(obs, action)
        with torch.no_grad():
            target_value = self.critic_targets(reward, next_obs, done)
        loss = loss_fn(value, target_value)

        stats = {"loss(critics)": loss.item()}
        stats.update(self.q_value_info(value))
        stats.update(self.model_dist_info(dist_params))
        return loss, stats
